import logging
from collections import Counter
from copy import copy
from pathlib import Path
from typing import Iterable, NamedTuple, Union

import fs
from fs import path as fspath
from fs.base import FS
from fs.errors import NoSysPath, ResourceNotFound
from fs.walk import Walker
from rich.console import Console

from . import config, console
from .actions import ACTIONS
from .actions.action import Action
from .filters import FILTERS
from .filters.filter import Filter
from .migration import migrate_v1
from .utils import (
    basic_args,
    deep_merge_inplace,
    ensure_dict,
    ensure_list,
    fs_path_from_options,
    to_args,
)

logger = logging.getLogger(__name__)
highlighted_console = Console()


class Location(NamedTuple):
    walker: Walker
    fs: FS
    fs_path: str


DEFAULT_SYSTEM_EXCLUDE_FILES = [
    "thumbs.db",
    "desktop.ini",
    "~$*",
    ".DS_Store",
    ".localized",
]

DEFAULT_SYSTEM_EXCLUDE_DIRS = [
    ".git",
    ".svn",
]


def convert_options_to_walker_args(options: dict):
    # combine system_exclude and exclude into a single list
    excludes = copy(
        ensure_list(options.get("system_exclude_files", DEFAULT_SYSTEM_EXCLUDE_FILES))
    )
    excludes.extend(ensure_list(options.get("exclude_files", [])))
    exclude_dirs = copy(
        ensure_list(options.get("system_exclude_dirs", DEFAULT_SYSTEM_EXCLUDE_DIRS))
    )
    exclude_dirs.extend(ensure_list(options.get("exclude_dirs", [])))

    if not excludes:
        excludes = None
    if not exclude_dirs:
        exclude_dirs = None

    filter_ = copy(ensure_list(options.get("filter", [])))
    filter_dirs = copy(ensure_list(options.get("filter_dirs", [])))

    if not filter_:
        filter_ = None
    if not filter_dirs:
        filter_dirs = None

    # return all the default options
    result = {
        "ignore_errors": options.get("ignore_errors", False),
        "on_error": options.get("on_error", None),
        "search": options.get("search", "depth"),
        "exclude": excludes,
        "exclude_dirs": exclude_dirs,
        "max_depth": options.get("max_depth", None),
        "filter": filter_,
        "filter_dirs": filter_dirs,
    }
    return result


def instantiate_location(
    options: Union[str, dict], default_filesystem, default_max_depth=0
) -> Location:
    if isinstance(options, Location):
        return options
    if isinstance(options, str):
        options = {"path": options}

    # set default max depth from rule
    if not "max_depth" in options:
        options["max_depth"] = default_max_depth

    if "walker" not in options:
        args = convert_options_to_walker_args(options)
        walker = Walker(**args)
    else:
        walker = options["walker"]

    fs, fs_path = fs_path_from_options(
        path=options.get("path", "."),
        filesystem=options.get("filesystem", default_filesystem),
    )
    return Location(walker=walker, fs=fs, fs_path=fs_path)


def instantiate_filter(filter_config):
    if isinstance(filter_config, Filter):
        return filter_config
    spec = ensure_dict(filter_config)
    name, value = next(iter(spec.items()))
    parts = name.split(maxsplit=1)
    invert = False
    if len(parts) == 2 and parts[0] == "not":
        name = parts[1]
        invert = True
    args, kwargs = to_args(value)
    instance = FILTERS[name](*args, **kwargs)
    instance.set_logic(inverted=invert)
    return instance


def instantiate_action(action_config):
    if isinstance(action_config, Action):
        return action_config
    spec = ensure_dict(action_config)
    name, value = next(iter(spec.items()))
    args, kwargs = to_args(value)
    return ACTIONS[name](*args, **kwargs)


def syspath_or_exception(fs, path):
    try:
        return Path(fs.getsyspath(path))
    except NoSysPath as e:
        return e


def replace_with_instances(config: dict, default_filesystem):
    warnings = []

    for rule in config["rules"]:
        default_depth = None if rule.get("subfolders", False) else 0

        _locations = []
        for options in ensure_list(rule["locations"]):
            try:
                instance = instantiate_location(
                    options=options,
                    default_max_depth=default_depth,
                    default_filesystem=default_filesystem,
                )
                _locations.append(instance)
            except Exception as e:
                if isinstance(options, dict) and options.get("ignore_errors", False):
                    warnings.append(str(e))
                else:
                    raise ValueError("Invalid location %s (%s)" % (options, e)) from e

        # filters are optional
        _filters = []
        for x in ensure_list(rule.get("filters", [])):
            try:
                _filters.append(instantiate_filter(x))
            except Exception as e:
                raise ValueError("Invalid filter %s (%s)" % (x, e)) from e

        # actions
        _actions = []
        for x in ensure_list(rule["actions"]):
            try:
                _actions.append(instantiate_action(x))
            except Exception as e:
                raise ValueError("Invalid action %s (%s)" % (x, e)) from e

        rule["locations"] = _locations
        rule["filters"] = _filters
        rule["actions"] = _actions

    return warnings


def filter_pipeline(filters: Iterable[Filter], args: dict, filter_mode: str) -> bool:
    """
    run the filter pipeline.
    Returns True on a match, False otherwise and updates `args` in the process.
    """
    results = []
    for filter_ in filters:
        try:
            # update dynamic path args
            args["fs_base_path"] = fs.path.abspath(
                fs.path.normpath(args["fs_base_path"])
            )
            args["fs_path"] = fs.path.abspath(fs.path.normpath(args["fs_path"]))
            args["path"] = syspath_or_exception(args["fs"], args["fs_path"])
            args["relative_path"] = fspath.frombase(
                args["fs_base_path"], args["fs_path"]
            )

            match, updates = filter_.pipeline(args)
            result = match ^ filter_.inverted
            # we cannot exit early on "any".
            if (filter_mode == "none" and result) or (
                filter_mode == "all" and not result
            ):
                return False
            results.append(result)
            deep_merge_inplace(args, updates)
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(e)
            # console.print_exception()
            filter_.print_error(str(e))
            return False

    if filter_mode == "any":
        return any(results)
    return True


def action_pipeline(actions: Iterable[Action], args: dict, simulate: bool) -> bool:
    for action in actions:
        try:
            # update dynamic path args
            args["fs_base_path"] = fs.path.abspath(
                fs.path.normpath(args["fs_base_path"])
            )
            args["fs_path"] = fs.path.abspath(fs.path.normpath(args["fs_path"]))
            args["path"] = syspath_or_exception(args["fs"], args["fs_path"])
            args["relative_path"] = fspath.frombase(
                args["fs_base_path"], args["fs_path"]
            )

            updates = action.pipeline(args, simulate=simulate)
            # jobs may return a dict with updates that should be merged into args
            if updates is not None:
                deep_merge_inplace(args, updates)
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(e)
            action.print_error(str(e))
            return False
    return True


def should_execute(rule_tags, tags, skip_tags):
    if not rule_tags:
        rule_tags = set()
    if not tags:
        tags = set()
    if not skip_tags:
        skip_tags = set()

    if "always" in rule_tags and "always" not in skip_tags:
        return True
    if "never" in rule_tags and "never" not in tags:
        return False
    if not tags and not skip_tags:
        return True
    if not rule_tags and tags:
        return False
    should_run = any(tag in tags for tag in rule_tags) or not tags or not rule_tags
    should_skip = any(tag in skip_tags for tag in rule_tags)
    return should_run and not should_skip


def run_rules(rules: dict, tags, skip_tags, simulate: bool = True):
    count = Counter(done=0, fail=0)  # type: Counter

    if simulate:
        console.simulation_banner()

    console.spinner(simulate=simulate)
    for rule_nr, rule in enumerate(rules["rules"], start=1):
        rule_tags = rule.get("tags")
        if isinstance(rule_tags, str):
            rule_tags = [tag.strip() for tag in rule_tags.split(",")]
        should_run = should_execute(
            rule_tags=rule_tags,
            tags=tags,
            skip_tags=skip_tags,
        )
        if not should_run:
            continue
        target = rule.get("targets", "files")
        console.rule(rule.get("name", "Rule %s" % rule_nr))
        filter_mode = rule.get("filter_mode", "all")

        for walker, walker_fs, walker_path in rule["locations"]:
            console.location(walker_fs, walker_path)
            walk = walker.files if target == "files" else walker.dirs
            for path in walk(fs=walker_fs, path=walker_path):
                try:
                    if walker_fs.islink(path):
                        continue
                except ResourceNotFound:
                    console.warn(
                        "Ignoring "
                        + walker_fs.getsyspath(path)
                        + " (may be a broken symlink)"
                    )
                    continue

                # tell the user which resource we're handling
                console.path(walker_fs, path)

                # assemble the available args
                args = basic_args()
                args.update(
                    fs=walker_fs,
                    fs_path=path,
                    fs_base_path=walker_path,
                )

                # run resource through the filter pipeline
                match = filter_pipeline(
                    filters=rule["filters"],
                    args=args,
                    filter_mode=filter_mode,
                )

                # if the currently handled resource changed we adjust the prefix message
                if args.get("resource_changed"):
                    console.path_changed_during_pipeline(
                        fs=walker_fs,
                        fs_path=path,
                        new_fs=args["fs"],
                        new_path=args["fs_path"],
                        reason=args.get("resource_changed"),
                    )
                args.pop("resource_changed", None)

                # run resource through the action pipeline
                if match:
                    is_success = action_pipeline(
                        actions=rule["actions"],
                        args=args,
                        simulate=simulate,
                    )
                    if is_success:
                        count["done"] += 1
                    else:
                        count["fail"] += 1

    if simulate:
        console.simulation_banner()

    return count


def run(
    rules: Union[str, dict],
    simulate: bool,
    working_dir: Union[FS, str] = ".",
    validate=True,
    tags=None,
    skip_tags=None,
):
    # load and validate
    if isinstance(rules, str):
        rules = config.load_from_string(rules)

    rules = config.cleanup(rules)
    Action.Meta.default_filesystem = working_dir

    migrate_v1(rules)

    if validate:
        config.validate(rules)

    # instantiate
    warnings = replace_with_instances(rules, default_filesystem=working_dir)
    for msg in warnings:
        console.warn(msg)

    # run
    count = run_rules(rules=rules, tags=tags, skip_tags=skip_tags, simulate=simulate)
    console.summary(count)

    if count["fail"]:
        raise RuntimeWarning("Some actions failed.")



from fs import open_fs
from fs.memoryfs import MemoryFS

from organize.utils import is_same_resource


def test_mem():
    a = open_fs("mem://")
    a.touch("file1")
    b = a.makedir("b")
    b.touch("file2")
    c = a

    assert is_same_resource(a, "b/file2", a, "b/file2")
    assert is_same_resource(a, "b/file2", b, "file2")
    assert is_same_resource(a, "file1", c, "file1")


def test_mem2():
    mem = MemoryFS()
    fs1, path1 = mem.makedir("files"), "test.txt"
    fs2, path2 = mem, "files/test.txt"

    assert is_same_resource(fs1, path1, fs2, path2)


def test_osfs():
    a = open_fs("~").makedir("Desktop", recreate=True)
    b = open_fs("~/")
    c = b.opendir("Desktop")

    assert is_same_resource(a, "file.txt", a, "file.txt")
    assert is_same_resource(a, "file.txt", b, "Desktop/file.txt")
    assert is_same_resource(a, "file.txt", c, "file.txt")


def test_inter():
    a = open_fs("temp://")
    b = open_fs(a.getsyspath("/"))
    a_dir = a.makedir("a")

    assert is_same_resource(a, "test.txt", b, "test.txt")
    assert is_same_resource(b, "a/subfile.txt", a_dir, "subfile.txt")
    # assert is_same_resource(a, "test.txt", a_dir, "../test.txt")


def test_nested():
    for protocol in ("mem://", "temp://"):
        with open_fs(protocol) as mem:
            x = mem.makedir("sub1")
            x = x.makedir("sub2")
            x = x.makedir("sub3")
            x.touch("file")

            y = mem.opendir("sub1")

            assert is_same_resource(mem, "sub1/sub2/sub3/file", x, "file")
            assert is_same_resource(y, "sub2/sub3/file", x, "file")



import operator
import re
from typing import Callable, Dict
from typing import Optional as Opt
from typing import Sequence, Set, Tuple

from fs.filesize import binary, decimal, traditional
from schema import Optional, Or

from organize.utils import flattened_string_list

from .filter import Filter, FilterResult

OPERATORS = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "=": operator.eq,
    "": operator.eq,
    ">=": operator.ge,
    ">": operator.gt,
}
SIZE_REGEX = re.compile(
    r"^(?P<op>[<>=]*)(?P<num>(\d*\.)?\d+)(?P<unit>[kmgtpezy]?i?)b?$"
)


def create_constraints(inp: str) -> Set[Tuple[Callable[[int, int], bool], int]]:
    """
    Given an input string it returns a list of tuples (comparison operator,
    number of bytes).

    Accepted formats are: "30k", ">= 5 TiB, <10tb", "< 60 tb", ...
    Calculation is in bytes, even if the "b" is lowercase. If an "i" is present
    we calculate base 1024.
    """
    result = set()  # type: Set[Tuple[Callable[[int, int], bool], int]]
    parts = str(inp).replace(" ", "").lower().split(",")
    for part in parts:
        try:
            reg_match = SIZE_REGEX.match(part)
            if reg_match:
                match = reg_match.groupdict()
                op = OPERATORS[match["op"]]
                num = float(match["num"]) if "." in match["num"] else int(match["num"])
                unit = match["unit"]
                base = 1024 if unit.endswith("i") else 1000
                exp = "kmgtpezy".index(unit[0]) + 1 if unit else 0
                numbytes = num * base**exp
                result.add((op, numbytes))
        except (AttributeError, KeyError, IndexError, ValueError, TypeError) as e:
            raise ValueError("Invalid size format: %s" % part) from e
    return result


def satisfies_constraints(size, constraints):
    return all(op(size, p_size) for op, p_size in constraints)


class Size(Filter):
    """Matches files and folders by size

    Args:
        *conditions (list(str) or str):
            The size constraints.

    Accepts file size conditions, e.g: `">= 500 MB"`, `"< 20k"`, `">0"`,
    `"= 10 KiB"`.

    It is possible to define both lower and upper conditions like this:
    `">20k, < 1 TB"`, `">= 20 Mb, <25 Mb"`. The filter will match if all given
    conditions are satisfied.

    - Accepts all units from KB to YB.
    - If no unit is given, kilobytes are assumend.
    - If binary prefix is given (KiB, GiB) the size is calculated using base 1024.

    **Returns:**

    - `{size.bytes}`: (int) Size in bytes
    - `{size.traditional}`: (str) Size with unit (powers of 1024, JDEC prefixes)
    - `{size.binary}`: (str) Size with unit (powers of 1024, IEC prefixes)
    - `{size.decimal}`: (str) Size with unit (powers of 1000, SI prefixes)
    """

    name = "size"
    arg_schema = Or(object, [object])
    schema_support_instance_without_args = True

    def __init__(self, *conditions: Sequence[str]) -> None:
        self.conditions = ", ".join(flattened_string_list(list(conditions)))
        self.constraints = create_constraints(self.conditions)

    def matches(self, filesize: int) -> bool:
        if not self.constraints:
            return True
        return all(op(filesize, c_size) for op, c_size in self.constraints)

    def pipeline(self, args: dict) -> FilterResult:
        fs = args["fs"]
        fs_path = args["fs_path"]

        if fs.isdir(fs_path):
            size = sum(
                info.size
                for _, info in fs.walk.info(path=fs_path, namespaces=["details"])
            )
        else:
            size = fs.getsize(fs_path)

        return FilterResult(
            matches=self.matches(size),
            updates={
                self.get_name(): {
                    "bytes": size,
                    "traditional": traditional(size),
                    "binary": binary(size),
                    "decimal": decimal(size),
                },
            },
        )

    def __str__(self) -> str:
        return "FileSize({})".format(" ".join(self.conditions))



from os.path import commonpath
from typing import Callable, Union

from fs import errors, open_fs
from fs.base import FS
from fs.copy import copy_file
from fs.errors import FSError
from fs.move import move_dir
from fs.opener import manage_fs
from fs.opener.errors import OpenerError
from fs.osfs import OSFS
from fs.path import dirname, frombase
from schema import Optional, Or

from organize.utils import SimulationFS, Template, safe_description

from .action import Action
from ._conflict_resolution import CONFLICT_OPTIONS, check_conflict, dst_from_options


# this is taken from my PR
def move_file_optimized(
    src_fs,
    src_path,
    dst_fs,
    dst_path,
    preserve_time=False,
    cleanup_dst_on_error=True,
):
    # type: (...) -> None
    """Move a file from one filesystem to another.

    Arguments:
        src_fs (FS or str): Source filesystem (instance or URL).
        src_path (str): Path to a file on ``src_fs``.
        dst_fs (FS or str): Destination filesystem (instance or URL).
        dst_path (str): Path to a file on ``dst_fs``.
        preserve_time (bool): If `True`, try to preserve mtime of the
            resources (defaults to `False`).
        cleanup_dst_on_error (bool): If `True`, tries to delete the file copied to
            ``dst_fs`` if deleting the file from ``src_fs`` fails (defaults to `True`).

    """
    with manage_fs(src_fs, writeable=True) as _src_fs:
        with manage_fs(dst_fs, writeable=True, create=True) as _dst_fs:
            if _src_fs is _dst_fs:
                # Same filesystem, may be optimized
                _src_fs.move(
                    src_path, dst_path, overwrite=True, preserve_time=preserve_time
                )
                return

            if _src_fs.hassyspath(src_path) and _dst_fs.hassyspath(dst_path):
                # if both filesystems have a syspath we create a new OSFS from a
                # common parent folder and use it to move the file.
                try:
                    src_syspath = _src_fs.getsyspath(src_path)
                    dst_syspath = _dst_fs.getsyspath(dst_path)
                    common = commonpath([src_syspath, dst_syspath])
                    if common:
                        rel_src = frombase(common, src_syspath)
                        rel_dst = frombase(common, dst_syspath)
                        with _src_fs.lock(), _dst_fs.lock():
                            with OSFS(common) as base:
                                base.move(rel_src, rel_dst, preserve_time=preserve_time)
                                return  # optimization worked, exit early
                except ValueError:
                    # This is raised if we cannot find a common base folder.
                    # In this case just fall through to the standard method.
                    pass

            # Standard copy and delete
            with _src_fs.lock(), _dst_fs.lock():
                copy_file(
                    _src_fs,
                    src_path,
                    _dst_fs,
                    dst_path,
                    preserve_time=preserve_time,
                )
                try:
                    _src_fs.remove(src_path)
                except FSError as e:
                    # if the source cannot be removed we delete the copy on the
                    # destination
                    if cleanup_dst_on_error:
                        _dst_fs.remove(dst_path)
                    raise e


class Move(Action):

    """Move a file to a new location.

    The file can also be renamed.
    If the specified path does not exist it will be created.

    If you only want to rename the file and keep the folder, it is
    easier to use the `rename` action.

    Args:
        dest (str):
            The destination where the file / dir should be moved to.
            If `dest` ends with a slash, it is assumed to be a target directory
            and the file / dir will be moved into `dest` and keep its name.

        on_conflict (str):
            What should happen in case **dest** already exists.
            One of `skip`, `overwrite`, `trash`, `rename_new` and `rename_existing`.
            Defaults to `rename_new`.

        rename_template (str):
            A template for renaming the file / dir in case of a conflict.
            Defaults to `{name} {counter}{extension}`.

        filesystem (str):
            (Optional) A pyfilesystem opener url of the filesystem you want to copy to.
            If this is not given, the local filesystem is used.

    The next action will work with the moved file / dir.
    """

    name = "move"
    arg_schema = Or(
        str,
        {
            "dest": str,
            Optional("on_conflict"): Or(*CONFLICT_OPTIONS),
            Optional("rename_template"): str,
            Optional("filesystem"): object,
        },
    )

    def __init__(
        self,
        dest: str,
        on_conflict="rename_new",
        rename_template="{name} {counter}{extension}",
        filesystem: Union[FS, str, None] = None,
    ) -> None:
        if on_conflict not in CONFLICT_OPTIONS:
            raise ValueError(
                "on_conflict must be one of %s" % ", ".join(CONFLICT_OPTIONS)
            )

        self.dest = Template.from_string(dest)
        self.conflict_mode = on_conflict
        self.rename_template = Template.from_string(rename_template)
        self.filesystem = filesystem or self.Meta.default_filesystem

    def pipeline(self, args: dict, simulate: bool):
        src_fs = args["fs"]  # type: FS
        src_path = args["fs_path"]

        move_action: Callable[[FS, str, FS, str], None]
        if src_fs.isdir(src_path):
            move_action = move_dir
        elif src_fs.isfile(src_path):
            move_action = move_file_optimized

        dst_fs, dst_path = dst_from_options(
            src_path=src_path,
            dest=self.dest,
            filesystem=self.filesystem,
            args=args,
        )

        # check for conflicts
        skip, dst_path = check_conflict(
            src_fs=src_fs,
            src_path=src_path,
            dst_fs=dst_fs,
            dst_path=dst_path,
            conflict_mode=self.conflict_mode,
            rename_template=self.rename_template,
            simulate=simulate,
            print=self.print,
        )

        try:
            dst_fs = open_fs(dst_fs, create=False, writeable=True)
        except (errors.CreateFailed, OpenerError):
            if not simulate:
                dst_fs = open_fs(dst_fs, create=True, writeable=True)
            else:
                dst_fs = SimulationFS(dst_fs)

        if not skip:
            self.print("Move to %s" % safe_description(dst_fs, dst_path))
            if not simulate:
                dst_fs.makedirs(dirname(dst_path), recreate=True)
                move_action(src_fs, src_path, dst_fs, dst_path)

        # the next action should work with the newly created copy
        return {
            "fs": dst_fs,
            "fs_path": dst_path,
        }

    def __str__(self) -> str:
        return "Move(dest=%s, conflict_mode=%s)" % (self.dest, self.conflict_mode)



import pytest
from conftest import make_files, read_files
from fs.base import FS

from organize import core

files = {
    "test.txt": "",
    "file.txt": "Hello world\nAnother line",
    "another.txt": "",
    "folder": {
        "x.txt": "",
    },
}


@pytest.fixture
def testfiles(testfs) -> FS:
    make_files(testfs, files)
    yield testfs


def test_copy_on_itself(testfiles):
    config = """
    rules:
      - locations: "/"
        actions:
          - move: "/"
    """
    core.run(config, simulate=True, working_dir=testfiles)
    result = read_files(testfiles)
    assert result == files


@pytest.mark.parametrize(
    "mode,files,test_txt_content",
    [
        ("skip", ["file.txt", "test.txt"], "old"),
        ("overwrite", ["test.txt"], "new"),
        ("rename_new", ["test.txt", "test 2.txt"], "old"),
        ("rename_existing", ["test.txt", "test 2.txt"], "new"),
    ],
)
def test_move_conflict(testfs: FS, mode, files, test_txt_content):
    config = """
    rules:
      - locations: "/"
        filters:
          - name: file
        actions:
          - move:
              dest: "test.txt"
              on_conflict: {}
    """.format(
        mode
    )
    testfs.writetext("file.txt", "new")
    testfs.writetext("test.txt", "old")
    core.run(config, simulate=False, working_dir=testfs)
    assert set(testfs.listdir("/")) == set(files)
    assert testfs.readtext("test.txt") == test_txt_content



import os
from pathlib import Path

from organize.actions import Move
from organize.utils import DotDict

USER_DIR = os.path.expanduser("~")

ARGS = DotDict(basedir=Path.home(), path=Path.home() / "test.py", simulate=False)


def test_tilde_expansion(mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir):
    mock_exists.return_value = False
    mock_samefile.return_value = False
    move = Move(dest="~/newname.py", overwrite=False)
    updates = move.run(**ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_move.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"), dst=os.path.join(USER_DIR, "newname.py")
    )
    assert updates == {"path": Path("~/newname.py").expanduser()}


def test_into_folder(mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir):
    mock_exists.return_value = False
    mock_samefile.return_value = False
    move = Move(dest="~/somefolder/", overwrite=False)
    updates = move.run(**ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_move.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "somefolder", "test.py"),
    )
    assert updates == {"path": Path(USER_DIR) / "somefolder" / "test.py"}


def test_overwrite(mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir):
    mock_exists.return_value = True
    mock_samefile.return_value = False
    move = Move(dest="~/somefolder/", overwrite=True)
    updates = move.run(**ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_called_with(os.path.join(USER_DIR, "somefolder", "test.py"))
    mock_move.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "somefolder", "test.py"),
    )
    assert updates is not None


def test_already_exists(mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir):
    mock_exists.side_effect = [True, False]
    mock_samefile.return_value = False
    move = Move(dest="~/folder/", overwrite=False)
    updates = move.run(**ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_move.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "folder", "test 2.py"),
    )
    assert updates is not None


def test_already_exists_multiple(
    mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir
):
    mock_exists.side_effect = [True, True, True, False]
    mock_samefile.return_value = False
    move = Move(dest="~/folder/", overwrite=False)
    updates = move.run(**ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_move.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "folder", "test 4.py"),
    )
    assert updates is not None


def test_already_exists_multiple_separator(
    mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir
):
    mock_exists.side_effect = [True, True, True, False]
    mock_samefile.return_value = False
    move = Move(dest="~/folder/", overwrite=False, counter_separator="_")
    updates = move.run(**ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_move.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "folder", "test_4.py"),
    )
    assert updates is not None


def test_makedirs(mock_parent, mock_move, mock_trash):
    move = Move(dest="~/some/new/folder/", overwrite=False)
    updates = move.run(**ARGS)
    mock_parent.mkdir.assert_called_with(parents=True, exist_ok=True)
    mock_trash.assert_not_called()
    mock_move.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "some", "new", "folder", "test.py"),
    )
    assert updates is not None


def test_args(mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir):
    args = ARGS.merge({"nr": {"upper": 1}})
    mock_exists.return_value = False
    mock_samefile.return_value = False
    move = Move(dest="~/{nr.upper}-name.py", overwrite=False)
    updates = move.run(**args)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_move.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"), dst=os.path.join(USER_DIR, "1-name.py")
    )
    assert updates is not None


def test_path(mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir):
    mock_exists.return_value = False
    mock_samefile.return_value = False
    move = Move(dest="~/{path.stem}/{path.suffix}/{path.name}", overwrite=False)
    updates = move.run(**ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_move.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "test", ".py", "test.py"),
    )
    assert updates is not None


def test_keep_location(mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir):
    mock_exists.return_value = True
    mock_samefile.return_value = True
    move = Move(dest="~/test.py")
    updates = move.run(**ARGS)
    mock_mkdir.assert_not_called()
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_move.assert_not_called()
    assert updates is not None


def test_dont_keep_case_sensitive(
    mock_exists, mock_samefile, mock_move, mock_trash, mock_mkdir
):
    mock_exists.return_value = True
    mock_samefile.return_value = True
    move = Move(dest="~/TEST.PY")
    updates = move.run(**ARGS)
    assert mock_mkdir.call_count > 0
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    assert mock_move.call_count > 0
    assert updates is not None



import logging

from fs.base import FS

from organize.utils import Template

from .filter import Filter, FilterResult

logger = logging.getLogger(__name__)


class Hash(Filter):

    """Calculates the hash of a file.

    Args:
        algorithm (str): Any hashing algorithm available to python's `hashlib`.
            `md5` by default.

    Algorithms guaranteed to be available are
    `shake_256`, `sha3_256`, `sha1`, `sha3_224`, `sha384`, `sha512`, `blake2b`,
    `blake2s`, `sha256`, `sha224`, `shake_128`, `sha3_512`, `sha3_384` and `md5`.

    Depending on your python installation and installed libs there may be additional
    hash algorithms to chose from.

    To list the available algorithms on your installation run this in a python
    interpreter:

    ```py
    >>> import hashlib
    >>> hashlib.algorithms_available
    {'shake_256', 'whirlpool', 'mdc2', 'blake2s', 'sha224', 'shake_128', 'sha3_512', 'sha3_224', 'sha384', 'md5', 'sha1', 'sha512_256', 'blake2b', 'sha256', 'sha512_224', 'ripemd160', 'sha3_384', 'md4', 'sm3', 'sha3_256', 'md5-sha1', 'sha512'}
    ```

    **Returns:**

    - `{hash}`:  The hash of the file.
    """

    name = "hash"
    schema_support_instance_without_args = True

    def __init__(self, algorithm="md5"):
        self.algorithm = Template.from_string(algorithm)

    def pipeline(self, args: dict):
        fs = args["fs"]  # type: FS
        fs_path = args["fs_path"]  # type: str
        algo = self.algorithm.render(**args)
        hash_ = fs.hash(fs_path, name=algo)
        return FilterResult(
            matches=True,
            updates={self.get_name(): hash_},
        )

    def __str__(self) -> str:
        return "Hash(algorithm={})".format(self.algorithm)



from conftest import make_files, read_files

from organize import core


def test_filecontent(tempfs):
    # inspired by https://github.com/tfeldmann/organize/issues/43
    files = {
        "Test1.txt": "Lorem MegaCorp Ltd. ipsum\nInvoice 12345\nMore text\nID: 98765",
        "Test2.txt": "Tests",
        "Test3.txt": "My Homework ...",
    }
    make_files(tempfs, files)
    config = r"""
        rules:
        - locations: "."
          filters:
            - filecontent: 'MegaCorp Ltd.+^Invoice (?P<number>\w+)$.+^ID: 98765$'
          actions:
            - rename: "MegaCorp_Invoice_{filecontent.number}.txt"
        - locations: "."
          filters:
            - filecontent: '.*Homework.*'
          actions:
            - rename: "Homework.txt"
        """
    core.run(config, simulate=False, working_dir=tempfs)
    assert read_files(tempfs) == {
        "Homework.txt": "My Homework ...",
        "MegaCorp_Invoice_12345.txt": "Lorem MegaCorp Ltd. ipsum\nInvoice 12345\nMore text\nID: 98765",
        "Test2.txt": "Tests",
    }



import fs
import pytest
from conftest import read_files

from organize.actions._conflict_resolution import (
    next_free_name,
    resolve_overwrite_conflict,
)
from organize.utils import Template


@pytest.mark.parametrize(
    "template,wanted,result",
    (
        ("{name}-{counter}{extension}", "file.txt", "file-2.txt"),
        ("{name}-{counter}{extension}", "file.txt", "file-2.txt"),
        ("{name}-{'%02d' % counter}{extension}", "file.txt", "file-03.txt"),
        ("{name}{counter}{extension}", "file.txt", "file4.txt"),
        ("{name}{counter}{extension}", "other.txt", "other.txt"),
    ),
)
def test_next_free_name(template, wanted, result):
    with fs.open_fs("mem://") as mem:
        mem.touch("file.txt")
        mem.touch("file1.txt")
        mem.touch("file-01.txt")
        mem.touch("file-02.txt")
        mem.touch("file2.txt")
        mem.touch("file3.txt")
        name, ext = fs.path.splitext(wanted)
        tmp = Template.from_string(template)
        assert next_free_name(mem, tmp, name, ext) == result


def test_next_free_name_exception():
    with fs.open_fs("mem://") as mem:
        mem.touch("file.txt")
        mem.touch("file1.txt")
        tmp = Template.from_string("{name}{extension}")
        with pytest.raises(ValueError):
            assert next_free_name(fs=mem, template=tmp, name="file", extension=".txt")


@pytest.mark.parametrize(
    "mode,result,files",
    (
        ("skip", None, {"file.txt": "file", "other.txt": "other"}),
        ("overwrite", "other.txt", {"file.txt": "file"}),
        # ("trash", None, {"file.txt": "file", "other.txt": "other"}),
        ("rename_new", "other2.txt", {"file.txt": "file", "other.txt": "other"}),
        ("rename_existing", "other.txt", {"file.txt": "file", "other2.txt": "other"}),
    ),
)
def test_resolve_overwrite_conflict(mode, result, files):
    with fs.open_fs("mem://") as mem:
        mem.writetext("file.txt", "file")
        mem.writetext("other.txt", "other")
        assert result == resolve_overwrite_conflict(
            src_fs=mem,
            src_path="file.txt",
            dst_fs=mem,
            dst_path="other.txt",
            conflict_mode=mode,
            rename_template=Template.from_string("{name}{counter}{extension}"),
            simulate=False,
            print=print,
        )
        assert files == read_files(mem)



import os
from pathlib import Path

from organize.actions import Copy

USER_DIR = os.path.expanduser("~")

DEFAULT_ARGS = {
    "basedir": Path.home(),
    "path": Path.home() / "test.py",
    "simulate": False,
}


def test_tilde_expansion(mock_exists, mock_samefile, mock_copy, mock_trash, mock_mkdir):
    mock_exists.return_value = False
    mock_samefile.return_value = False
    copy = Copy(dest="~/newname.py", overwrite=False)
    updates = copy.run(**DEFAULT_ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_copy.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"), dst=os.path.join(USER_DIR, "newname.py")
    )
    # keep old file path
    assert updates is None


def test_into_folder(mock_exists, mock_samefile, mock_copy, mock_trash, mock_mkdir):
    mock_exists.return_value = False
    mock_samefile.return_value = False
    copy = Copy(dest="~/somefolder/", overwrite=False)
    copy.run(**DEFAULT_ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_copy.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "somefolder", "test.py"),
    )


def test_overwrite(mock_exists, mock_samefile, mock_copy, mock_trash, mock_mkdir):
    mock_exists.return_value = True
    mock_samefile.return_value = False
    copy = Copy(dest="~/somefolder/", overwrite=True)
    copy.run(**DEFAULT_ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_called_with(os.path.join(USER_DIR, "somefolder", "test.py"))
    mock_copy.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "somefolder", "test.py"),
    )


def test_already_exists_multiple_with_separator(
    mock_exists, mock_samefile, mock_copy, mock_trash, mock_mkdir
):
    args = {
        "basedir": Path.home(),
        "path": Path.home() / "test_2.py",
        "simulate": False,
    }
    mock_exists.side_effect = [True, True, True, False]
    mock_samefile.return_value = False
    copy = Copy(dest="~/folder/", overwrite=False, counter_separator="_")
    copy.run(**args)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_copy.assert_called_with(
        src=os.path.join(USER_DIR, "test_2.py"),
        dst=os.path.join(USER_DIR, "folder", "test_5.py"),
    )


def test_makedirs(mock_parent, mock_copy, mock_trash):
    copy = Copy(dest="~/some/new/folder/", overwrite=False)
    copy.run(**DEFAULT_ARGS)
    mock_parent.mkdir.assert_called_with(parents=True, exist_ok=True)
    mock_trash.assert_not_called()
    mock_copy.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "some", "new", "folder", "test.py"),
    )


def test_args(mock_exists, mock_samefile, mock_copy, mock_trash, mock_mkdir):
    args = {
        "basedir": Path.home(),
        "path": Path.home() / "test.py",
        "simulate": False,
        "nr": {"upper": 1},
    }
    mock_exists.return_value = False
    mock_samefile.return_value = False
    copy = Copy(dest="~/{nr.upper}-name.py", overwrite=False)
    copy.run(**args)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_copy.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"), dst=os.path.join(USER_DIR, "1-name.py")
    )


def test_path(mock_exists, mock_samefile, mock_copy, mock_trash, mock_mkdir):
    mock_exists.return_value = False
    mock_samefile.return_value = False
    copy = Copy(dest="~/{path.stem}/{path.suffix}/{path.name}", overwrite=False)
    copy.run(**DEFAULT_ARGS)
    mock_mkdir.assert_called_with(exist_ok=True, parents=True)
    mock_exists.assert_called_with()
    mock_trash.assert_not_called()
    mock_copy.assert_called_with(
        src=os.path.join(USER_DIR, "test.py"),
        dst=os.path.join(USER_DIR, "test", ".py", "test.py"),
    )



from datetime import datetime
from unittest.mock import patch

from organize.actions import Echo


def test_echo_basic():
    echo = Echo("Hello World")
    with patch.object(echo, "print") as m:
        echo.run(simulate=False)
        m.assert_called_with("Hello World")


def test_echo_args():
    echo = Echo('Date formatting: {now.strftime("%Y-%m-%d")}')
    with patch.object(echo, "print") as m:
        echo.run(simulate=False, now=datetime(2019, 1, 5))
        m.assert_called_with("Date formatting: 2019-01-05")


def test_echo_path():
    echo = Echo("{year}")
    with patch.object(echo, "print") as m:
        echo.run(simulate=False, year=2017)
        m.assert_called_with("2017")



import mimetypes

from organize.utils import flatten

from .filter import Filter, FilterResult


class MimeType(Filter):

    """Filter by MIME type associated with the file extension.

    Supports a single string or list of MIME type strings as argument.
    The types don't need to be fully specified, for example "audio" matches everything
    from "audio/midi" to "audio/quicktime".

    You can see a list of known MIME types on your system by running this oneliner:

    ```sh
    python3 -c "import mimetypes as m; print('\\n'.join(sorted(set(m.common_types.values()) | set(m.types_map.values()))))"
    ```

    Args:
        *mimetypes (list(str) or str): The MIME types to filter for.

    **Returns:**

    - `{mimetype}`: The MIME type of the file.
    """

    name = "mimetype"
    schema_support_instance_without_args = True

    def __init__(self, *mimetypes):
        self.mimetypes = list(map(str.lower, flatten(list(mimetypes))))

    @staticmethod
    def mimetype(path):
        type_, _ = mimetypes.guess_type(path, strict=False)
        return type_

    def matches(self, mimetype) -> bool:
        if mimetype is None:
            return False
        if not self.mimetypes:
            return True
        return any(mimetype.startswith(x) for x in self.mimetypes)

    def pipeline(self, args: dict) -> FilterResult:
        fs = args["fs"]
        fs_path = args["fs_path"]
        if fs.isdir(fs_path):
            raise ValueError("Dirs not supported.")
        mimetype = self.mimetype(fs_path)
        return FilterResult(
            matches=self.matches(mimetype),
            updates={self.get_name(): mimetype},
        )

    def __str__(self):
        return "MimeType(%s)" % ", ".join(self.mimetypes)



import pytest
from conftest import make_files, read_files

from organize import core


@pytest.mark.parametrize("loc", (".", "/"))
def test_rename_files_date(testfs, loc):
    # inspired by https://github.com/tfeldmann/organize/issues/43
    files = {
        "File_abc_dat20190812_xyz.pdf": "",
        "File_xyz_bar19990101_a.pdf": "",
        "File_123456_foo20000101_xyz20190101.pdf": "",
    }
    make_files(testfs, files)
    config = (
        r"""
    rules:
    - locations: "%s"
      filters:
        - regex: 'File_.*?(?P<y>\d{4})(?P<m>\d{2})(?P<d>\d{2}).*?.pdf'
      actions:
        - rename: "File_{regex.d}{regex.m}{regex.y}.pdf"
    """
        % loc
    )
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {
        "File_12082019.pdf": "",
        "File_01011999.pdf": "",
        "File_01012000.pdf": "",
    }



import subprocess
from datetime import datetime, timezone

from fs.base import FS

from ._timefilter import TimeFilter


class Created(TimeFilter):
    """Matches files / folders by created date

    Args:
        years (int): specify number of years
        months (int): specify number of months
        weeks (float): specify number of weeks
        days (float): specify number of days
        hours (float): specify number of hours
        minutes (float): specify number of minutes
        seconds (float): specify number of seconds
        mode (str):
            either 'older' or 'newer'. 'older' matches files / folders created before the given
            time, 'newer' matches files / folders created within the given time.
            (default = 'older')

    Returns:
        {created}: The datetime the file / folder was created.
    """

    name = "created"

    def get_datetime(self, args) -> datetime:
        fs = args["fs"]  # type: FS
        fs_path = args["fs_path"]
        created = fs.getinfo(fs_path, namespaces=["details"]).created
        if not created:
            # We're probably on Linux. No easy way to get creation dates here,
            # so we try to use the stat utility.
            created = self.fallback_method(fs, fs_path)
        if not created:
            raise EnvironmentError("File creation time is not available.")
        return created

    def fallback_method(self, fs, fs_path):
        if fs.hassyspath(fs_path):
            syspath = fs.getsyspath(fs_path)
            commands = (
                ["stat", "--format=%W", syspath],  # GNU coreutils
                ["stat", "-f %B", syspath],  # BSD
            )
            for cmd in commands:
                try:
                    created_str = subprocess.check_output(cmd, encoding="utf-8").strip()
                    timestamp = int(created_str)
                    return datetime.fromtimestamp(timestamp, tz=timezone.utc)
                except subprocess.CalledProcessError:
                    pass

    def __str__(self):
        return "<Created>"



import pathlib
from typing import Union
from unittest.mock import patch

import fs
import pytest
from fs.base import FS
from fs.path import basename, join

ORGANIZE_DIR = pathlib.Path(__file__).parent.parent


@pytest.fixture
def tempfs():
    with fs.open_fs("temp://") as tmp:
        yield tmp


@pytest.fixture
def memfs():
    with fs.open_fs("mem://") as mem:
        yield mem


@pytest.fixture(
    params=[
        "mem://",
        pytest.param("temp://"),
    ],
)
def testfs(request) -> FS:
    with fs.open_fs(request.param) as tmp:
        yield tmp


@pytest.fixture
def mock_echo():
    with patch("organize.actions.Echo.print") as mck:
        yield mck


def make_files(fs: FS, layout: Union[dict, list], path="/"):
    """
    Example layout:

        layout = {
            "folder": {
                "subfolder": {
                    "test.txt": "",
                    "other.pdf": b"binary",
                },
            },
            "file.txt": "Hello world\nAnother line",
        }
    """
    fs.makedirs(path, recreate=True)
    if isinstance(layout, list):
        for f in layout:
            fs.touch(f)
        return
    for k, v in layout.items():
        respath = join(path, k)

        # folders are dicts
        if isinstance(v, dict):
            make_files(fs=fs, layout=v, path=respath)

        # everything else is a file
        elif v is None:
            fs.touch(respath)
        elif isinstance(v, bytes):
            fs.writebytes(respath, v)
        elif isinstance(v, str):
            fs.writetext(respath, v)
        else:
            raise ValueError("Unknown file data %s" % v)


def read_files(fs: FS, path="/"):
    result = dict()
    for x in fs.walk.files(path, max_depth=0):
        result[basename(x)] = fs.readtext(x)
    for x in fs.walk.dirs(path, max_depth=0):
        result[basename(x)] = read_files(fs, path=join(path, x))
    return result



import logging
from typing import Any, Dict
from typing import Optional as tyOptional
from typing import Union

from fs.base import FS
from schema import Optional, Or, Schema

from organize.console import pipeline_error, pipeline_message

logger = logging.getLogger(__name__)


class Error(Exception):
    pass


class Action:
    name = None  # type: Union[str, None]
    arg_schema = None
    schema_support_instance_without_args = False

    class Meta:
        default_filesystem = "." # type: Union[FS, str]

    @classmethod
    def get_name(cls):
        if cls.name:
            return cls.name
        return cls.__name__.lower()

    @classmethod
    def get_schema(cls):
        if cls.arg_schema:
            arg_schema = cls.arg_schema
        else:
            arg_schema = Or(
                str,
                [str],
                Schema({}, ignore_extra_keys=True),
            )
        if cls.schema_support_instance_without_args:
            return Or(
                cls.get_name(),
                {
                    cls.get_name(): arg_schema,
                },
            )
        return {
            cls.get_name(): arg_schema,
        }

    def run(self, simulate: bool, **kwargs) -> tyOptional[Dict[str, Any]]:
        return self.pipeline(kwargs, simulate=simulate)

    def pipeline(self, args: dict, simulate: bool) -> tyOptional[Dict[str, Any]]:
        raise NotImplementedError

    def print(self, *msg) -> None:
        """print a message for the user"""
        text = " ".join(str(x) for x in msg)
        for line in text.splitlines():
            pipeline_message(source=self.get_name(), msg=line)

    def print_error(self, msg: str):
        for line in msg.splitlines():
            pipeline_error(source=self.get_name(), msg=line)

    def __str__(self) -> str:
        return self.__class__.__name__

    def __repr__(self) -> str:
        return "<%s>" % str(self)

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__



import textwrap
from collections import defaultdict

import yaml
from schema import And, Literal, Optional, Or, Schema

from organize.actions import ACTIONS
from organize.filters import FILTERS

from .utils import flatten_all_lists_in_dict

CONFIG_SCHEMA = Schema(
    {
        Literal("rules", description="All rules are defined here."): [
            {
                Optional("name", description="The name of the rule."): str,
                Optional("enabled"): bool,
                Optional("subfolders"): bool,
                Optional("filter_mode", description="The filter mode."): Or(
                    "all", "any", "none", error="Invalid filter mode"
                ),
                Optional("tags"): Or(str, [str]),
                Optional(
                    "targets",
                    description="Whether the rule should apply to directories or folders.",
                ): Or("dirs", "files"),
                "locations": Or(
                    str,
                    [
                        Or(
                            str,
                            {
                                "path": And(str, len),
                                Optional("max_depth"): Or(int, None),
                                Optional("search"): Or("depth", "breadth"),
                                Optional("exclude_files"): Or(str, [str]),
                                Optional("exclude_dirs"): Or(str, [str]),
                                Optional("system_exclude_files"): Or(str, [str]),
                                Optional("system_exclude_dirs"): Or(str, [str]),
                                Optional("filter"): Or(str, [str]),
                                Optional("filter_dirs"): Or(str, [str]),
                                Optional("ignore_errors"): bool,
                                Optional("filesystem"): object,
                            },
                        ),
                    ],
                ),
                Optional("filters"): [
                    Optional(x.get_schema()) for x in FILTERS.values()
                ],
                "actions": [Optional(x.get_schema()) for x in ACTIONS.values()],
            },
        ],
    },
    name="organize rule configuration",
)


def default_yaml_cnst(loader, tag_suffix, node):
    # disable yaml constructors for strings starting with exclamation marks
    # https://stackoverflow.com/a/13281292/300783
    return str(node.tag)


yaml.add_multi_constructor("", default_yaml_cnst, Loader=yaml.SafeLoader)


def load_from_string(config: str) -> dict:
    dedented_config = textwrap.dedent(config)
    return yaml.load(dedented_config, Loader=yaml.SafeLoader)


def lowercase_keys(obj):
    if isinstance(obj, dict):
        obj = {key.lower(): value for key, value in obj.items()}
        for key, value in obj.items():
            if isinstance(value, list):
                for i, item in enumerate(value):
                    value[i] = lowercase_keys(item)
            obj[key] = lowercase_keys(value)
    return obj


def cleanup(config: dict) -> dict:
    result = defaultdict(list)

    # delete every root key except "rules"
    for rule in config.get("rules", []):
        # delete disabled rules
        if rule.get("enabled", True):
            result["rules"].append(rule)

    if not result:
        raise ValueError("No rules defined.")

    result = lowercase_keys(result)

    # flatten all lists everywhere
    return flatten_all_lists_in_dict(dict(result))


def validate(config: dict):
    return CONFIG_SCHEMA.validate(config)



from datetime import datetime, timedelta

from organize.filters import Created


def test_min():
    now = datetime.now()
    ct = Created(days=10, hours=12, mode="older")
    assert not ct.matches_datetime(now - timedelta(days=10, hours=0))
    assert ct.matches_datetime(now - timedelta(days=10, hours=13))


def test_max():
    now = datetime.now()
    ct = Created(days=10, hours=12, mode="newer")
    assert ct.matches_datetime(now - timedelta(days=10, hours=0))
    assert not ct.matches_datetime(now - timedelta(days=10, hours=13))



from pathlib import Path
from unittest.mock import patch

from organize.actions import Python


def test_print_substitution():
    with patch.object(Python, "print") as mock_print:
        python = Python("print('Hello World')")
        python.run(path=Path.home(), simulate=False)
        mock_print.assert_called_with("Hello World")


def test_code_execution():
    with patch.object(Python, "print") as mock_print:
        path = Path("/some/folder")
        python = Python("print(x)\nprint(path)")
        python.run(path=path, x=42, simulate=False)
        mock_print.assert_any_call(42)
        mock_print.assert_any_call(path)



import textwrap
from typing import Any
from typing import Optional as tyOpt
from typing import Sequence

from schema import Or

from .filter import Filter, FilterResult


class Python(Filter):

    """Use python code to filter files.

    Args:
        code (str):
            The python code to execute. The code must contain a `return` statement.


    **Returns:**

    - If your code returns `False` or `None` the file is filtered out,
        otherwise the file is passed on to the next filters.
    - `{python}` contains the returned value. If you return a dictionary (for
        example `return {"some_key": some_value, "nested": {"k": 2}}`) it will be
        accessible via dot syntax in your actions: `{python.some_key}`,
        `{python.nested.k}`.
    """

    name = "python"

    arg_schema = Or(str, {"code": str})

    def __init__(self, code) -> None:
        self.code = textwrap.dedent(code)
        if "return" not in self.code:
            raise ValueError("No return statement found in your code!")

    def usercode(self, *args, **kwargs) -> tyOpt[Any]:
        pass  # will be overwritten by `create_method`

    def create_method(self, name: str, argnames: Sequence[str], code: str) -> None:
        globals_ = globals().copy()
        globals_["print"] = self.print
        locals_ = locals().copy()
        locals_["self"] = self
        funccode = "def {fnc}__({arg}):\n{cod}\n\nself.{fnc} = {fnc}__\n".format(
            fnc=name,
            arg=", ".join(argnames),
            cod=textwrap.indent(textwrap.dedent(code), " " * 4),
        )
        exec(funccode, globals_, locals_)  # pylint: disable=exec-used

    def pipeline(self, args) -> FilterResult:
        self.create_method(name="usercode", argnames=args.keys(), code=self.code)
        result = self.usercode(**args)  # pylint: disable=assignment-from-no-return
        if result not in (False, None):
            return FilterResult(matches=True, updates={self.get_name(): result})
        return FilterResult(matches=False, updates={})



import logging
import subprocess
from subprocess import PIPE

from schema import Optional, Or

from ..utils import Template
from .action import Action

logger = logging.getLogger(__name__)


class Shell(Action):
    """
    Executes a shell command

    Args:
        cmd (str): The command to execute.
        run_in_simulation (bool):
            Whether to execute in simulation mode (default = false)
        ignore_errors (bool):
            Whether to continue on returncodes != 0.
        simulation_output (str):
            The value of `{shell.output}` if run in simulation
        simulation_returncode (int):
            The value of `{shell.returncode}` if run in simulation

    Returns

    - `{shell.output}` (`str`): The stdout of the executed process.
    - `{shell.returncode}` (`int`): The returncode of the executed process.
    """

    name = "shell"
    arg_schema = Or(
        str,
        {
            "cmd": str,
            Optional("run_in_simulation"): bool,
            Optional("ignore_errors"): bool,
            Optional("simulation_output"): str,
            Optional("simulation_returncode"): int,
        },
    )

    def __init__(
        self,
        cmd: str,
        run_in_simulation=False,
        ignore_errors=False,
        simulation_output="** simulation **",
        simulation_returncode=0,
    ):
        self.cmd = Template.from_string(cmd)
        self.run_in_simulation = run_in_simulation
        self.ignore_errors = ignore_errors
        self.simulation_output = Template.from_string(simulation_output)
        self.simulation_returncode = simulation_returncode

    def pipeline(self, args: dict, simulate: bool):
        full_cmd = self.cmd.render(**args)

        if not simulate or self.run_in_simulation:
            # we use call instead of run to be compatible with python < 3.5
            self.print("$ %s" % full_cmd)
            logger.info('Executing command "%s" in shell.', full_cmd)
            try:
                call = subprocess.run(
                    full_cmd,
                    check=True,
                    stdout=PIPE,
                    stderr=subprocess.STDOUT,
                    shell=True,
                )
                return {
                    self.get_name(): {
                        "output": call.stdout.decode("utf-8"),
                        "returncode": 0,
                    }
                }
            except subprocess.CalledProcessError as e:
                if not self.ignore_errors:
                    raise e
                return {
                    self.get_name(): {
                        "output": e.stdout.decode("utf-8"),
                        "returncode": e.returncode,
                    }
                }
        else:
            self.print("** not run in simulation ** $ %s" % full_cmd)
            return {
                self.get_name(): {
                    "output": self.simulation_output.render(**args),
                    "returncode": self.simulation_returncode,
                }
            }

    def __str__(self) -> str:
        return 'Shell(cmd="%s")' % self.cmd



from unittest.mock import call

from conftest import make_files
from fs import open_fs
from fs.path import dirname

from organize import core
from organize.filters import Extension


def test_extension():
    extension = Extension("JPG", ".gif", "pdf")
    testpathes = [
        ("/somefile.pdf", True),
        ("/home/test/somefile.pdf.jpeg", False),
        ("/home/test/gif.TXT", False),
        ("/home/test/txt.GIF", True),
        ("/somefile.pdf", True),
    ]
    with open_fs("mem://", writeable=True, create=True) as mem:
        for f, match in testpathes:
            mem.makedirs(dirname(f), recreate=True)
            mem.touch(f)
            assert extension.run(fs=mem, fs_path=f).matches == match


def test_extension_empty():
    with open_fs("mem://") as mem:
        mem.touch("test.txt")
        extension = Extension()
        assert extension.run(fs=mem, fs_path="test.txt").matches


def test_extension_result():
    with open_fs("mem://") as mem:

        path = "somefile.TxT"
        mem.touch(path)

        extension = Extension("txt")
        assert extension.matches(".TxT")
        result = extension.run(fs=mem, fs_path=path).updates["extension"]
        assert str(result) == "TxT"
        assert result.lower() == "txt"
        assert result.upper() == "TXT"

        extension = Extension(".txt")
        assert extension.matches(".TXT")
        result = extension.run(fs=mem, fs_path=path).updates["extension"]
        assert str(result) == "TxT"
        assert result.lower() == "txt"
        assert result.upper() == "TXT"


def test_filename_move(testfs, mock_echo):
    files = {
        "test.jpg": "",
        "asd.JPG": "",
        "nomatch.jpg.zip": "",
        "camel.jPeG": "",
    }
    make_files(testfs, files)
    config = """
        rules:
        - locations: "."
          filters:
            - name
            - extension:
              - .jpg
              - jpeg
          actions:
            - echo: 'Found JPG file: {name}'
        """
    core.run(config, simulate=False, working_dir=testfs)
    mock_echo.assert_has_calls(
        (
            call("Found JPG file: test"),
            call("Found JPG file: asd"),
            call("Found JPG file: camel"),
        ),
        any_order=True,
    )



import logging

from fs.base import FS

from organize.utils import safe_description

from .action import Action

logger = logging.getLogger(__name__)


class Delete(Action):

    """
    Delete a file from disk.

    Deleted files have no recovery option!
    Using the `Trash` action is strongly advised for most use-cases!
    """

    name = "delete"

    @classmethod
    def get_schema(cls):
        return cls.name

    def pipeline(self, args: dict, simulate: bool):
        fs = args["fs"]  # type: FS
        fs_path = args["fs_path"]  # type: str
        desc = safe_description(fs=fs, path=fs_path)
        self.print('Deleting "%s"' % desc)
        if not simulate:
            logger.info("Deleting %s.", desc)
            if fs.isdir(fs_path):
                fs.removetree(fs_path)
            elif fs.isfile(fs_path):
                fs.remove(fs_path)



import argparse
import getpass
import os
import re
import subprocess
from datetime import datetime
from pathlib import Path

import requests

SRC_FOLDER = "organize"
CURRENT_FOLDER = Path(__file__).resolve().parent
GITHUB_API_ENDPOINT = "https://api.github.com/repos/tfeldmann/organize"


def ask_confirm(text):
    while True:
        answer = input(f"{text} [y/n]: ").lower()
        if answer in ("j", "y", "ja", "yes"):
            return True
        if answer in ("n", "no", "nein"):
            return False


def set_version(args):
    """
    - reads and validates version number
    - updates __version__.py
    - updates pyproject.toml
    - Searches for 'WIP' in changelog and replaces it with current version and date
    """
    from organize.__version__ import __version__ as current_version

    print(f"Current version is {current_version}.")

    # read version from input if not given
    version = args.version

    if not version:
        version = input("Version number: ")

    # validate and remove 'v' if present
    version = version.lower()
    if not re.match(r"v?\d+\.\d+.*", version):
        return
    if version.startswith("v"):
        version = version[1:]

    # safety check
    if not ask_confirm(f"Creating version v{version}. Continue?"):
        return

    # update library version
    versionfile = CURRENT_FOLDER / SRC_FOLDER / "__version__.py"
    with open(versionfile, "w") as f:
        print(f"Updating {versionfile}")
        f.write(f'__version__ = "{version}"\n')

    # update poetry version
    print("Updating pyproject.toml")
    subprocess.run(["poetry", "version", version], check=True)

    # read changelog
    print("Updating CHANGELOG.md")
    with open(CURRENT_FOLDER / "CHANGELOG.md", "r") as f:
        changelog = f.read()

    # check if WIP section is in changelog
    wip_regex = re.compile(r"## WIP\n(.*?)(?=\n##)", re.MULTILINE | re.DOTALL)
    match = wip_regex.search(changelog)
    if not match:
        print('No "## WIP" section found in changelog')
        return

    # change WIP to version number and date
    changes = match.group(1)
    today = datetime.now().strftime("%Y-%m-%d")
    changelog = wip_regex.sub(f"## v{version} ({today})\n{changes}", changelog, count=1)

    # write changelog
    with open(CURRENT_FOLDER / "CHANGELOG.md", "w") as f:
        f.write(changelog)

    if ask_confirm("Commit changes?"):
        subprocess.run(
            ["git", "add", "pyproject.toml", "*/__version__.py", "CHANGELOG.md"]
        )
        subprocess.run(["git", "commit", "-m", f"bump version to v{version}"])

    print("Please push to github and wait for CI to pass.")
    print("Success.")


def publish(args):
    """
    - reads version
    - reads changes from changelog
    - creates git tag
    - pushes to github
    - publishes on pypi
    - creates github release
    """
    from organize.__version__ import __version__ as version

    if not ask_confirm(f"Publishing version {version}. Is this correct?"):
        return

    if ask_confirm("Run the tests?"):
        os.system("poetry run pytest")
        os.system("poetry run mypy organize main.py")

    # extract changes from changelog
    with open(CURRENT_FOLDER / "CHANGELOG.md", "r") as f:
        changelog = f.read()
    wip_regex = re.compile(
        "## v{}".format(version.replace(".", r"\.")) + r".*?\n(.*?)(?=\n##)",
        re.MULTILINE | re.DOTALL,
    )
    match = wip_regex.search(changelog)
    if not match:
        print("Failed to extract changes from changelog. Do the versions match?")
        return
    changes = match.group(1).strip()
    print(f"Changes:\n{changes}")

    # create git tag ('vXXX')
    if ask_confirm("Create tag?"):
        subprocess.run(["git", "tag", "-a", f"v{version}", "-m", f"v{version}"])

    # push to github
    if ask_confirm("Push to github?"):
        print("Pushing to github")
        subprocess.run(["git", "push", "--follow-tags"], check=True)

    # upload to pypi
    if ask_confirm("Publish on Pypi?"):
        subprocess.run(["rm", "-rf", "dist"], check=True)
        subprocess.run(["poetry", "build"], check=True)
        subprocess.run(["poetry", "publish"], check=True)

    # create github release
    if ask_confirm("Create github release?"):
        response = requests.post(
            f"{GITHUB_API_ENDPOINT}/releases",
            auth=(input("Benutzer: "), getpass.getpass(prompt="API token: ")),
            json={
                "tag_name": f"v{version}",
                "target_commitish": "main",
                "name": f"v{version}",
                "body": changes,
                "draft": False,
                "prerelease": False,
            },
        )
        response.raise_for_status()

    print("Success.")


def main():
    assert CURRENT_FOLDER == Path.cwd().resolve()

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    parser_version = subparsers.add_parser("version", help="Set the version number")
    parser_version.add_argument(
        "version", type=str, help="The version number", nargs="?", default=None
    )
    parser_version.set_defaults(func=set_version)

    parser_publish = subparsers.add_parser("publish", help="Publish the project")
    parser_publish.set_defaults(func=publish)

    args = parser.parse_args()
    if not vars(args):
        parser.print_help()
    else:
        args.func(args)


if __name__ == "__main__":
    main()



import pytest
from conftest import make_files, read_files

from organize import core
from organize.filters import Size


def test_constrains_mope1():
    assert not Size("<1b,>2b").matches(1)
    assert Size(">=1b,<2b").matches(1)
    assert not Size(">1.000001b").matches(1)
    assert Size("<1.000001B").matches(1)
    assert Size("<1.000001").matches(1)
    assert Size("<=1,>=0.001kb").matches(1)
    assert Size("<1").matches(0)
    assert not Size(">1").matches(0)
    assert not Size("<1,>1b").matches(0)
    assert Size(">99.99999GB").matches(100000000000)
    assert Size("0").matches(0)


def test_constrains_base():
    assert Size(">1kb,<1kib").matches(1010)
    assert Size(">1k,<1ki").matches(1010)
    assert Size("1k").matches(1000)
    assert Size("1000").matches(1000)


def test_other():
    assert Size("<100 Mb").matches(20)
    assert Size("<100 Mb, <10 mb, <1 mb, > 0").matches(20)
    assert Size(["<100 Mb", ">= 0 Tb"]).matches(20)


def test_size_zero(testfs):
    files = {"1": "", "2": "", "3": ""}
    make_files(testfs, files)
    config = """
        rules:
        - locations: "."
          filters:
            - size: 0
          actions:
            - echo: '{path.name}'
            - delete
        """
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {}


def test_basic(testfs):
    files = {
        "empty": "",
        "full": "0" * 2000,
        "halffull": "0" * 1010,
        "two_thirds.txt": "0" * 666,
    }
    make_files(testfs, files)
    config = """
        rules:
        - locations: "."
          filters:
            - size: '> 1kb, <= 1.0 KiB'
          actions:
            - echo: '{path.name} {size.bytes}'
        - locations: "."
          filters:
            - not size:
              - '> 0.5 kb'
              - '<1.0 KiB'
          actions:
            - delete
        """
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {
        "halffull": "0" * 1010,
        "two_thirds.txt": "0" * 666,
    }


@pytest.mark.skip(reason="TODO - template vars in filters not supported")
def test_python_args(testfs):
    make_files(
        testfs,
        {
            "empty": "",
            "full": "0" * 2000,
            "halffull": "0" * 1010,
            "two_thirds.txt": "0" * 666,
        },
    )
    config = """
    rules:
    - folders: files
        filters:
        - python: |
            return 2000
        - filesize: '= {python}b'
        actions:
        - delete
    """
    core.run(config=config, working_dir=testfs, simulate=False)
    assert read_files(testfs) == {
        "empty": "",
        "halffull": "0" * 1010,
        "two_thirds.txt": "0" * 666,
    }



import logging
import sys

import simplematch as sm
from schema import Or

from organize.utils import Template

from .action import Action

logger = logging.getLogger(__name__)


class MacOSTags(Action):

    """Add macOS tags.

    Args:
        *tags (str): A list of tags or a single tag.

    The color can be specified in brackets after the tag name, for example:

    ```yaml
    macos_tags: "Invoices (red)"
    ```

    Available colors are `none`, `gray`, `green`, `purple`, `blue`, `yellow`, `red` and
    `orange`.
    """

    name = "macos_tags"

    @classmethod
    def get_schema(cls):
        return {cls.name: Or(str, [str])}

    def __init__(self, *tags):
        self.tags = [Template.from_string(tag) for tag in tags]

    def pipeline(self, args: dict, simulate: bool):
        fs = args["fs"]
        fs_path = args["fs_path"]
        path = fs.getsyspath(fs_path)

        if sys.platform != "darwin":
            raise EnvironmentError("The macos_tags action is only available on macOS")

        import macos_tags

        COLORS = [c.name.lower() for c in macos_tags.Color]

        for template in self.tags:
            tag = template.render(**args)
            name, color = self._parse_tag(tag)

            if color not in COLORS:
                raise ValueError(
                    "color %s is unknown. (Available: %s)" % (color, " / ".join(COLORS))
                )

            self.print('Adding tag: "%s" (color: %s)' % (name, color))
            if not simulate:
                _tag = macos_tags.Tag(
                    name=name,
                    color=macos_tags.Color[color.upper()],
                )  # type: ignore
                macos_tags.add(_tag, file=str(path))

    def _parse_tag(self, s):
        """parse a tag definition and return a tuple (name, color)"""
        result = sm.match("{name} ({color})", s)
        if not result:
            return s, "none"
        return result["name"], result["color"].lower()

    def __str__(self) -> str:
        return "MacOSTags(tags=%s)" % self.tags



import sys
import pytest
from fs import open_fs

from organize.filters import MacOSTags


@pytest.mark.skipif(sys.platform != "darwin", reason="runs only on macOS")
def test_macos_filter():
    import macos_tags

    tags = MacOSTags("Invoice (red)", "* (green)")
    with open_fs("temp://", writeable=True, create=True) as temp:
        temp.touch("My-Invoice.pdf")
        path = temp.getsyspath("My-Invoice.pdf")
        macos_tags.add(macos_tags.Tag("Invoice", macos_tags.Color.RED), file=path)

        temp.touch("Another-File.pdf")
        path = temp.getsyspath("Another-File.pdf")
        macos_tags.add(macos_tags.Tag("Urgent", macos_tags.Color.GREEN), file=path)

        temp.touch("Pic.jpg")
        path = temp.getsyspath("Pic.jpg")
        macos_tags.add(macos_tags.Tag("Pictures", macos_tags.Color.BLUE), file=path)

        assert tags.run(fs=temp, fs_path="My-Invoice.pdf").matches
        assert tags.run(fs=temp, fs_path="Another-File.pdf").matches
        assert not tags.run(fs=temp, fs_path="Pic.jpg").matches


def test_macos_tags_matching():
    tags = MacOSTags("Invoice (*)", "* (red)", "Test (green)")
    assert tags.matches(["Invoice (none)"])
    assert tags.matches(["Invoice (green)"])
    assert not tags.matches(["Voice (green)"])
    assert tags.matches(["Voice (red)"])
    assert not tags.matches(["Test (blue)"])
    assert tags.matches(["Test (green)"])



from typing import Any, Dict, List, Union

import simplematch
from fs import path
from schema import Optional, Or

from .filter import Filter, FilterResult


class Name(Filter):
    """Match files and folders by name

    Args:
        match (str):
            A matching string in [simplematch-syntax](https://github.com/tfeldmann/simplematch)

        startswith (str):
            The filename must begin with the given string

        contains (str):
            The filename must contain the given string

        endswith (str):
            The filename (without extension) must end with the given string

        case_sensitive (bool):
            By default, the matching is case sensitive. Change this to False to use
            case insensitive matching.
    """

    name = "name"
    schema_support_instance_without_args = True

    arg_schema = Or(
        str,
        {
            Optional("match"): str,
            Optional("startswith"): Or(str, [str]),
            Optional("contains"): Or(str, [str]),
            Optional("endswith"): Or(str, [str]),
            Optional("case_sensitive"): bool,
        },
    )

    def __init__(
        self,
        match="*",
        *,
        startswith="",
        contains="",
        endswith="",
        case_sensitive=True,
    ) -> None:
        self.matcher = simplematch.Matcher(match, case_sensitive=case_sensitive)
        self.startswith = self.create_list(startswith, case_sensitive)
        self.contains = self.create_list(contains, case_sensitive)
        self.endswith = self.create_list(endswith, case_sensitive)
        self.case_sensitive = case_sensitive

    def matches(self, name: str) -> bool:
        if not self.case_sensitive:
            name = name.lower()

        is_match = (
            self.matcher.test(name)
            and any(x in name for x in self.contains)
            and any(name.startswith(x) for x in self.startswith)
            and any(name.endswith(x) for x in self.endswith)
        )
        return is_match

    def pipeline(self, args: Dict) -> FilterResult:
        fs = args["fs"]
        fs_path = args["fs_path"]
        if fs.isdir(fs_path):
            name = path.basename(fs_path)
        else:
            name, ext = path.splitext(path.basename(fs_path))
            if not name:
                name = ext
        result = self.matches(name)
        m = self.matcher.match(name)
        if m == {}:
            m = name
        return FilterResult(
            matches=result,
            updates={self.get_name(): m},
        )

    @staticmethod
    def create_list(x: Union[int, str, List[Any]], case_sensitive: bool) -> List[str]:
        if isinstance(x, (int, float)):
            x = str(x)
        if isinstance(x, str):
            x = [x]
        x = [str(x) for x in x]
        if not case_sensitive:
            x = [x.lower() for x in x]
        return x



MIGRATION_DOCS_URL = "https://organize.readthedocs.io/en/latest/updating-from-v1/"


class MigrationWarning(UserWarning):
    pass


class NeedsMigrationError(Exception):
    pass


def entry_name_args(entry):
    if isinstance(entry, str):
        return (entry.lower(), [])
    elif isinstance(entry, dict):
        name, value = next(iter(entry.items()))
        if isinstance(value, str):
            return (name.lower(), [])
        elif isinstance(value, dict):
            args = [x.lower() for x in value.keys()]
            return (name.lower(), args)
        return (name.lower(), [])


def migrate_v1(config: dict):
    for rule in config.get("rules") or []:
        if "folders" in rule:
            raise NeedsMigrationError("`folders` are now `locations`")
        for fil in rule.get("filters") or []:
            name, _ = entry_name_args(fil)
            if name == "filename":
                raise NeedsMigrationError("`filename` is now `name`")
            if name == "filesize":
                raise NeedsMigrationError("`filesize` is now `size`")
        for act in rule.get("actions") or []:
            if act:
                name, args = entry_name_args(act)
                if name in ("move", "copy", "rename"):
                    if "overwrite" in args or "counter_seperator" in args:
                        raise NeedsMigrationError(
                            "`%s` does not support `overwrite` and "
                            "`counter_seperator` anymore. Please use the new arguments."
                            % name
                        )



from conftest import make_files, read_files

from organize import core


def test_dependent_rules(testfs):
    files = {
        "asd.txt": "",
        "newname 2.pdf": "",
        "newname.pdf": "",
        "test.pdf": "",
    }
    make_files(testfs, files)
    config = """
    rules:
    - locations: "."
      filters:
        - name: test
      actions:
        - copy: newfolder/test.pdf
    - locations: "newfolder"
      filters:
        - name: test
      actions:
        - rename: test-found.pdf
    """
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {
        "newname.pdf": "",
        "newname 2.pdf": "",
        "test.pdf": "",
        "asd.txt": "",
        "newfolder": {"test-found.pdf": ""},
    }



import logging
import logging.config

import yaml
from fs import appfs

with appfs.UserLogFS("organize") as log_fs:
    LOG_PATH = log_fs.getsyspath("organize.log")

# configure logging
LOGGING_CONFIG = """
version: 1
disable_existing_loggers: false
formatters:
    simple:
        format: '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
handlers:
    console:
        class: logging.StreamHandler
        level: DEBUG
        formatter: simple
        stream: ext://sys.stdout
    file:
        class: logging.handlers.TimedRotatingFileHandler
        level: DEBUG
        filename: {filename}
        formatter: simple
        when: midnight
        backupCount: 30
root:
    level: DEBUG
    handlers: [file]
exifread:
    level: INFO
""".format(
    filename=str(LOG_PATH)
)
logging.config.dictConfig(yaml.safe_load(LOGGING_CONFIG))



"""
Duplicate detection filter.

Based on this stackoverflow answer:
    https://stackoverflow.com/a/36113168/300783

Which I updated for python3 in:
    https://gist.github.com/tfeldmann/fc875e6630d11f2256e746f67a09c1ae
"""
import hashlib
from collections import defaultdict
from typing import Dict, NamedTuple, Set, Union

from fs.base import FS
from fs.errors import NoSysPath, NoURL
from fs.path import basename

from organize.utils import is_same_resource

from .filter import Filter, FilterResult

HASH_ALGORITHM = "sha1"
DETECTION_METHODS = ("first_seen", "name", "created", "lastmodified")
DETECTION_METHOD_REGEX = r"(-?)\s*?({})".format("|".join(DETECTION_METHODS))


class File(NamedTuple):
    fs: FS
    path: str
    base_path: str

    @property
    def lastmodified(self):
        return self.fs.getmodified(self.path)

    @property
    def created(self):
        return self.fs.getinfo(self.path, namespaces=["details"]).created

    @property
    def name(self):
        return basename(self.path)

    @property
    def ident(self):
        try:
            return self.fs.getsyspath(self.path)
        except NoSysPath:
            pass
        try:
            return self.fs.geturl(self.path)
        except NoURL:
            pass
        return "%s,%s" % (self.path, id(self.fs))


def getsize(f: File):
    return f.fs.getsize(f.path)


def full_hash(f: File):
    return f.fs.hash(f.path, name=HASH_ALGORITHM)


def first_chunk_hash(f: File):
    hash_object = hashlib.new(HASH_ALGORITHM)
    with f.fs.openbin(f.path) as file_:
        hash_object.update(file_.read(1024))
    return hash_object.hexdigest()


def detect_original(known: File, new: File, method: str, reverse: bool):
    """Returns a tuple (original file, duplicate)"""

    if method == "first_seen":
        return (known, new) if not reverse else (new, known)
    elif method == "name":
        return tuple(sorted((known, new), key=lambda x: x.name, reverse=reverse))
    elif method == "created":
        return tuple(sorted((known, new), key=lambda x: x.created, reverse=reverse))
    elif method == "lastmodified":
        return tuple(
            sorted((known, new), key=lambda x: x.lastmodified, reverse=reverse)
        )
    else:
        raise ValueError("Unknown original detection method: %s" % method)


class Duplicate(Filter):
    """A fast duplicate file finder.

    This filter compares files byte by byte and finds identical files with potentially
    different filenames.

    Args:
        detect_original_by (str):
            Detection method to distinguish between original and duplicate.
            Possible values are:

            - `"first_seen"`: Whatever file is visited first is the original. This
              depends on the order of your location entries.
            - `"name"`: The first entry sorted by name is the original.
            - `"created"`: The first entry sorted by creation date is the original.
            - `"lastmodified"`: The first file sorted by date of last modification is the original.

    You can reverse the sorting method by prefixing a `-`.

    So with `detect_original_by: "-created"` the file with the older creation date is
    the original and the younger file is the duplicate. This works on all methods, for
    example `"-first_seen"`, `"-name"`, `"-created"`, `"-lastmodified"`.

    **Returns:**

    `{duplicate.original}` - The path to the original
    """

    name = "duplicate"
    schema_support_instance_without_args = True

    def __init__(self, detect_original_by="first_seen"):
        if detect_original_by.startswith("-"):
            self.detect_original_by = detect_original_by[1:]
            self.select_orignal_reverse = True
        else:
            self.detect_original_by = detect_original_by
            self.select_orignal_reverse = False

        self.files_for_size = defaultdict(list)
        self.files_for_chunk = defaultdict(list)
        self.file_for_hash = dict()

        # we keep track of the files we already computed the hashes for so we only do
        # that once.
        self.seen_files = set()  # type: Set[File]
        self.first_chunk_known = set()  # type: Set[File]
        self.hash_known = set()  # type: Set[File]

    def matches(self, fs: FS, path: str, base_path: str):
        file_ = File(fs=fs, path=path, base_path=base_path)

        # skip symlinks
        if fs.islink(path):
            return False

        # the exact same path has already been handled. This happens if multiple
        # locations emit this file in a single rule or if we follow symlinks.
        # We skip these.
        if file_ in self.seen_files or any(
            is_same_resource(file_.fs, file_.path, x.fs, x.path)
            for x in self.seen_files
        ):
            return False

        self.seen_files.add(file_)

        # check for files with equal size
        file_size = getsize(file_)
        same_size = self.files_for_size[file_size]
        same_size.append(file_)
        if len(same_size) == 1:
            # the file is unique in size and cannot be a duplicate
            return False

        # for all other files with the same file size:
        # make sure we know their hash of their first 1024 byte chunk
        for f in same_size[:-1]:
            if f not in self.first_chunk_known:
                chunk_hash = first_chunk_hash(f)
                self.first_chunk_known.add(f)
                self.files_for_chunk[chunk_hash].append(f)

        # check first chunk hash collisions with the current file
        chunk_hash = first_chunk_hash(file_)
        same_first_chunk = self.files_for_chunk[chunk_hash]
        same_first_chunk.append(file_)
        self.first_chunk_known.add(file_)
        if len(same_first_chunk) == 1:
            # the file has a unique small hash and cannot be a duplicate
            return False

        # Ensure we know the full hashes of all files with the same first chunk as
        # the investigated file
        for f in same_first_chunk[:-1]:
            if f not in self.hash_known:
                hash_ = full_hash(f)
                self.hash_known.add(f)
                self.file_for_hash[hash_] = f

        # check full hash collisions with the current file
        hash_ = full_hash(file_)
        self.hash_known.add(file_)
        known = self.file_for_hash.get(hash_)
        if known:
            original, duplicate = detect_original(
                known=known,
                new=file_,
                method=self.detect_original_by,
                reverse=self.select_orignal_reverse,
            )
            if known != original:
                self.file_for_hash[hash_] = original

            resource_changed_reason = "duplicate of" if known != original else None
            from organize.core import syspath_or_exception

            return {
                "fs": duplicate.fs,
                "fs_path": duplicate.path,
                "fs_base_path": duplicate.base_path,
                "resource_changed": resource_changed_reason,
                self.get_name(): {
                    "original": syspath_or_exception(original.fs, original.path)
                },
            }

        return False

    def pipeline(self, args):
        fs = args["fs"]
        fs_path = args["fs_path"]
        fs_base_path = args["fs_base_path"]
        if fs.isdir(fs_path):
            raise EnvironmentError("Dirs are not supported")
        result = self.matches(fs=fs, path=fs_path, base_path=fs_base_path)
        if result is False:
            return FilterResult(matches=False, updates={})
        return FilterResult(matches=True, updates=result)

    def __str__(self) -> str:
        return "Duplicate()"



import pytest
from organize.core import should_execute


@pytest.mark.parametrize(
    "result, rule_tags, tags, skip_tags",
    (
        # no tags given
        (True, None, None, None),
        (True, ["tag"], None, None),
        (True, ["tag", "tag2"], None, None),
        # run tagged
        (False, None, ["tag"], None),
        (True, ["tag"], ["tag"], None),
        (True, ["tag", "tag2"], ["tag"], None),
        (False, ["foo", "bar"], ["tag"], None),
        (False, ["taggity"], ["tag"], None),
        # skip
        (True, None, None, ["tag"]),
        (True, ["tag"], None, ["asd"]),
        (False, ["tag"], None, ["tag"]),
        (False, ["tag", "tag2"], None, ["tag"]),
        # combination
        (False, None, ["tag"], ["foo"]),
        (False, ["foo", "tag"], ["tag"], ["foo"]),
        # always
        (True, ["always"], ["debug", "test"], None),
        (True, ["always", "tag"], None, ["tag"]),
        # skip only if specifically requested
        (False, ["always", "tag"], None, ["always"]),
        # never
        (False, ["never"], None, None),
        (False, ["never", "tag"], ["tag"], None),
        # run only if specifically requested
        (True, ["never", "tag"], ["never"], None),
    ),
)
def test_tags(result, rule_tags, tags, skip_tags):
    assert should_execute(rule_tags=rule_tags, tags=tags, skip_tags=skip_tags) == result



from datetime import datetime

from fs.base import FS

from ._timefilter import TimeFilter


class LastModified(TimeFilter):

    """Matches files by last modified date

    Args:
        years (int): specify number of years
        months (int): specify number of months
        weeks (float): specify number of weeks
        days (float): specify number of days
        hours (float): specify number of hours
        minutes (float): specify number of minutes
        seconds (float): specify number of seconds
        mode (str):
            either 'older' or 'newer'. 'older' matches files / folders last modified before
            the given time, 'newer' matches files / folders last modified within the given
            time. (default = 'older')

    Returns:
        {lastmodified}: The datetime the files / folders was lastmodified.
    """

    name = "lastmodified"

    def get_datetime(self, args: dict) -> datetime:
        fs = args["fs"]  # type: FS
        fs_path = args["fs_path"]
        modified = fs.getmodified(fs_path)
        if not modified:
            raise EnvironmentError("lastmodified date is not available")
        return modified

    def __str__(self):
        return "[LastModified] All files / folders last modified %s than %s" % (
            self._mode,
            self.timedelta,
        )



from pathlib import Path

from organize import core
from organize.filters import Regex

TESTDATA = [
    ("RG123456123456-sig.pdf", True, "123456123456"),
    ("RG002312321542-sig.pdf", True, "002312321542"),
    ("RG002312321542.pdf", False, None),
]


def test_regex_backslash():
    regex = Regex(r"^\.pdf$")
    assert regex.matches(".pdf")
    assert not regex.matches("+pdf")
    assert not regex.matches("/pdf")
    assert not regex.matches("\\pdf")


def test_regex_basic():
    regex = Regex(r"^RG(\d{12})-sig\.pdf$")
    for path, match, _ in TESTDATA:
        assert bool(regex.matches(path)) == match


def test_regex_return():
    regex = Regex(r"^RG(?P<the_number>\d{12})-sig\.pdf$")
    for path, valid, test_result in TESTDATA:
        if valid:
            result = regex.run(fs_path=path)
            assert result.updates == {"regex": {"the_number": test_result}}
            assert result.matches == True


def test_regex_umlaut():
    regex = Regex(r"^Erträgnisaufstellung-(?P<year>\d*)\.pdf")
    doc = "Erträgnisaufstellung-1998.pdf"
    assert regex.matches(doc)
    result = regex.run(fs_path=doc)
    assert result.updates == {"regex": {"year": "1998"}}
    assert result.matches


def test_multiple_regex_placeholders(testfs):
    config = """
    rules:
      - locations: "."
        filters:
          - regex: (?P<word>\w+)-(?P<number>\d+).*
          - regex: (?P<all>.+?)\.\w{3}
          - extension
        actions:
          - write:
               text: '{regex.word} {regex.number} {regex.all} {extension}'
               path: out.txt
    """
    testfs.touch("test-123.jpg")
    testfs.touch("other-456.pdf")
    core.run(config, simulate=False, working_dir=testfs)
    out = testfs.readtext("out.txt")
    assert "test 123 test-123 jpg" in out
    assert "other 456 other-456 pdf" in out



from organize.utils import deep_merge, deep_merge_inplace


def test_merges_dicts():
    a = {"a": 1, "b": {"b1": 2, "b2": 3}}
    b = {"a": 1, "b": {"b1": 4}}

    print(deep_merge(a, b))
    assert deep_merge(a, b)["a"] == 1
    assert deep_merge(a, b)["b"]["b2"] == 3
    assert deep_merge(a, b)["b"]["b1"] == 4


def test_returns_copy():
    a = {"regex": {"first": "A", "second": "B"}}
    b = {"regex": {"third": "C"}}

    x = deep_merge(a, b)
    a["regex"]["first"] = "X"
    assert x["regex"]["first"] == "A"
    assert x["regex"]["second"] == "B"
    assert x["regex"]["third"] == "C"


def test_inserts_new_keys():
    """Will it insert new keys by default?"""
    a = {"a": 1, "b": {"b1": 2, "b2": 3}}
    b = {"a": 1, "b": {"b1": 4, "b3": 5}, "c": 6}

    assert deep_merge(a, b)["a"] == 1
    assert deep_merge(a, b)["b"]["b2"] == 3
    assert deep_merge(a, b)["b"]["b1"] == 4
    assert deep_merge(a, b)["b"]["b3"] == 5
    assert deep_merge(a, b)["c"] == 6


def test_does_not_insert_new_keys():
    """Will it avoid inserting new keys when required?"""
    a = {"a": 1, "b": {"b1": 2, "b2": 3}}
    b = {"a": 1, "b": {"b1": 4, "b3": 5}, "c": 6}

    assert deep_merge(a, b, add_keys=True) == {
        "a": 1,
        "b": {"b1": 4, "b2": 3, "b3": 5},
        "c": 6,
    }
    assert deep_merge(a, b, add_keys=False) == {
        "a": 1,
        "b": {"b1": 4, "b2": 3},
    }


def test_inplace_merge():
    a = {}
    b = {1: {2: 2, 3: 3, 4: {5: "fin."}}}
    a = deep_merge(a, b)
    assert a == b
    b[1][2] = 5
    assert a != b

    deep_merge_inplace(a, {1: {4: {5: "new.", 6: "fin."}, 2: "x"}})
    assert a == {1: {2: "x", 3: 3, 4: {5: "new.", 6: "fin."}}}


def test_inplace_keeptype():
    a = {}
    deep_merge_inplace(a, {"nr": {"upper": 1}})
    assert a["nr"]["upper"] == 1



import os
from pathlib import Path

from organize.actions import Rename

USER_DIR = os.path.expanduser("~")

ARGS = {"basedir": Path.home(), "path": Path.home() / "test.py", "simulate": False}


def test_tilde_expansion(mock_exists, mock_samefile, mock_rename, mock_trash):
    mock_exists.return_value = False
    mock_samefile.return_value = False
    rename = Rename(name="newname.py", overwrite=False)
    new_path = rename.run(**ARGS)
    assert mock_exists.call_count > 0
    mock_trash.assert_not_called()
    expected_path = (Path.home() / "newname.py").expanduser()
    mock_rename.assert_called_with(expected_path)
    assert new_path == {"path": expected_path}


def test_overwrite(mock_exists, mock_samefile, mock_rename, mock_trash):
    mock_exists.return_value = True
    mock_samefile.return_value = False
    rename = Rename(name="{path.stem} Kopie.py", overwrite=True)
    new_path = rename.run(**ARGS)
    assert mock_exists.call_count > 0
    mock_trash.assert_called_with(os.path.join(USER_DIR, "test Kopie.py"))
    mock_rename.assert_called_with(Path("~/test Kopie.py").expanduser())
    assert new_path is not None


def test_already_exists(mock_exists, mock_samefile, mock_rename, mock_trash):
    mock_exists.side_effect = [True, False]
    mock_samefile.return_value = False
    rename = Rename(name="asd.txt", overwrite=False)
    new_path = rename.run(**ARGS)
    assert mock_exists.call_count > 0
    mock_trash.assert_not_called()
    mock_rename.assert_called_with(Path("~/asd 2.txt").expanduser())
    assert new_path is not None


def test_overwrite_samefile(mock_exists, mock_samefile, mock_rename, mock_trash):
    args = {"basedir": Path.home(), "path": Path.home() / "test.PDF", "simulate": False}
    mock_exists.return_value = True
    mock_samefile.return_value = True
    rename = Rename(name="{path.stem}.pdf", overwrite=False)
    new_path = rename.run(**args)
    assert mock_exists.call_count > 0
    mock_trash.assert_not_called()
    mock_rename.assert_called_with((Path.home() / "test.pdf").expanduser())
    assert new_path is not None


def test_keep_name(mock_exists, mock_samefile, mock_rename, mock_trash):
    args = {"basedir": Path.home(), "path": Path.home() / "test.pdf", "simulate": False}
    mock_exists.return_value = True
    mock_samefile.return_value = True
    rename = Rename(name="{path.stem}.pdf", overwrite=False)
    new_path = rename.run(**args)
    assert mock_exists.call_count > 0
    mock_trash.assert_not_called()
    mock_rename.assert_not_called()
    assert new_path is not None


def test_already_exists_multiple(mock_exists, mock_samefile, mock_rename, mock_trash):
    mock_exists.side_effect = [True, True, True, False]
    mock_samefile.return_value = False
    rename = Rename(name="asd.txt", overwrite=False)
    new_path = rename.run(**ARGS)
    assert mock_exists.call_count > 0
    mock_trash.assert_not_called()
    mock_rename.assert_called_with(Path("~/asd 4.txt").expanduser())
    assert new_path is not None


def test_already_exists_multiple_separator(
    mock_exists, mock_samefile, mock_rename, mock_trash
):
    mock_exists.side_effect = [True, True, True, False]
    mock_samefile.return_value = False
    rename = Rename(name="asd.txt", overwrite=False, counter_separator="-")
    new_path = rename.run(**ARGS)
    assert mock_exists.call_count > 0
    mock_trash.assert_not_called()
    mock_rename.assert_called_with(Path("~/asd-4.txt").expanduser())
    assert new_path is not None


def test_args(mock_exists, mock_samefile, mock_rename, mock_trash):
    args = {
        "basedir": Path.home(),
        "path": Path.home() / "test.py",
        "nr": {"upper": 1},
        "simulate": False,
    }

    mock_exists.return_value = False
    mock_samefile.return_value = False
    rename = Rename(name="{nr.upper}-{path.stem} Kopie.py")
    new_path = rename.run(**args)
    assert mock_exists.call_count > 0
    mock_trash.assert_not_called()
    mock_rename.assert_called_with(Path("~/1-test Kopie.py").expanduser())
    assert new_path is not None



from pathlib import Path

from fs.base import FS
from fs.path import basename, dirname, forcedir, relpath
from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm as RichConfirm
from rich.prompt import Prompt as RichPrompt
from rich.status import Status
from rich.text import Text
from rich.theme import Theme

from organize.__version__ import __version__

from .utils import safe_description

ICON_DIR = "🗁"
ICON_FILE = ""
INDENT = " " * 2

theme = Theme(
    {
        "info": "dim cyan",
        "warning": "yellow",
        "error": "bold red",
        "simulation": "bold green",
        "status": "bold green",
        "rule": "bold cyan",
        "location.fs": "yellow",
        "location.base": "green",
        "location.main": "bold green",
        "path.base": "dim green",
        "path.main": "green",
        "path.icon": "green",
        "pipeline.source": "cyan",
        "pipeline.msg": "white",
        "pipeline.error": "bold red",
        "pipeline.prompt": "bold yellow",
        "summary.done": "bold green",
        "summary.fail": "red",
    }
)
console = Console(theme=theme, highlight=False)
status = Status("", console=console)


class Prefixer:
    def __init__(self):
        self.reset()

    def reset(self):
        self._args = None
        self._kwargs = None

    def set_prefix(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def ensure_prefix(self):
        if self._args is not None:
            console.print(*self._args, **self._kwargs)
            self.reset()

    def print(self, *args, **kwargs):
        self.ensure_prefix()
        console.print(*args, **kwargs)


with_path = Prefixer()
with_newline = Prefixer()


class PipelineMixin:
    @classmethod
    def set_source(cls, source):
        cls.validate_error_message = Text.assemble(
            _pipeline_base(source),
            ("Please enter Y or N", "prompt.invalid"),
        )

    def pre_prompt(self):
        with_path.ensure_prefix()


class Confirm(PipelineMixin, RichConfirm):
    pass


class Prompt(PipelineMixin, RichPrompt):
    pass


def _highlight_path(path, base_style, main_style, relative=False):
    dir_ = forcedir(dirname(path))
    if relative:
        dir_ = relpath(dir_)
    name = basename(path)
    return Text.assemble(
        (dir_, base_style),
        (name, main_style),
    )


def info(config: str, working_dir: str):
    console.print("organize {}".format(__version__))
    console.print('Config: "{}"'.format(config))
    if working_dir != Path("."):
        console.print('Working dir: "{}"'.format(working_dir))


def warn(msg, title="Warning"):
    console.print("[warning][b]{}:[/b] {}[/warning]".format(title, msg))


def deprecated(msg):
    warn(msg, title="Deprecated")


def error(msg, title="Error"):
    console.print("[error]{}: {}[/error]".format(title, msg))


def simulation_banner():
    console.print()
    console.print(Panel("SIMULATION", style="simulation"))


def spinner(simulate: bool):
    status_verb = "simulating" if simulate else "organizing"
    status.update(Text(status_verb, "status"))
    status.start()


def rule(rule):
    console.print()
    console.rule("[rule]:gear: %s" % rule, align="left", style="rule")
    with_newline.reset()


def location(fs: FS, fs_path: str):
    result = Text()
    if fs.hassyspath(fs_path):
        syspath = fs.getsyspath(fs_path)
        result = _highlight_path(syspath.rstrip("/"), "location.base", "location.main")
    else:
        result = Text.assemble(
            (str(fs), "location.fs"),
            " ",
            _highlight_path(fs_path.rstrip("/"), "location.base", "location.main"),
        )
    with_newline.print(result)


def path(fs: FS, fs_path: str):
    icon = ICON_DIR if fs.isdir(fs_path) else ICON_FILE
    msg = Text.assemble(
        INDENT,
        _highlight_path(fs_path, "path.base", "path.main", relative=True),
        " ",
        (icon, "path.icon"),
    )
    with_path.set_prefix(msg)


def path_changed_during_pipeline(
    fs: FS, fs_path: str, new_fs: FS, new_path: str, reason="deferred from"
):
    icon = ICON_DIR if new_fs.isdir(new_path) else ICON_FILE
    msg = Text.assemble(
        INDENT,
        _highlight_path(
            safe_description(new_fs, new_path), "path.base", "path.main", relative=True
        ),
        (" <- %s " % reason, "yellow"),
        _highlight_path(fs_path, "path.base", "path.main", relative=True),
        " ",
        (icon, "path.icon"),
    )
    with_path.set_prefix(msg)


def _pipeline_base(source: str):
    return Text.assemble(
        INDENT * 2,
        ("- ({}) ".format(source), "pipeline.source"),
    )


def pipeline_message(source: str, msg: str) -> None:
    line = Text.assemble(
        _pipeline_base(source),
        (msg, "pipeline.msg"),
    )
    with_path.print(line)
    with_newline.set_prefix("")


def pipeline_error(source: str, msg: str):
    line = _pipeline_base(source)
    line.append("ERROR! {}".format(msg), "pipeline.error")
    with_path.print(line)
    with_newline.set_prefix("")


def pipeline_confirm(source: str, msg: str, default: bool):
    status.stop()
    line = _pipeline_base(source)
    line.append(msg, "pipeline.prompt")
    Confirm.set_source(source)
    result = Confirm.ask(
        line,
        console=console,
        default=default,
    )
    with_newline.set_prefix("")
    status.start()
    return result


def summary(count: dict):
    status.stop()
    console.print()
    if not sum(count.values()):
        console.print("Nothing to do.")
    else:
        result = Text.assemble(
            ("success {done}".format(**count), "summary.done"),
            " / ",
            ("fail {fail}".format(**count), "summary.fail"),
        )
        console.print(result)



import logging
import os
from typing import Callable

from fs import path
from fs.base import FS
from fs.move import move_dir, move_file
from schema import Optional, Or

from organize.utils import Template, safe_description

from .action import Action
from ._conflict_resolution import CONFLICT_OPTIONS, check_conflict, resolve_overwrite_conflict

logger = logging.getLogger(__name__)


class Rename(Action):

    """Renames a file.

    Args:
        name (str):
            The new name for the file / dir.

        on_conflict (str):
            What should happen in case **dest** already exists.
            One of `skip`, `overwrite`, `trash`, `rename_new` and `rename_existing`.
            Defaults to `rename_new`.

        rename_template (str):
            A template for renaming the file / dir in case of a conflict.
            Defaults to `{name} {counter}{extension}`.

    The next action will work with the renamed file / dir.
    """

    name = "rename"
    arg_schema = Or(
        str,
        {
            "name": str,
            Optional("on_conflict"): Or(*CONFLICT_OPTIONS),
            Optional("rename_template"): str,
        },
    )

    def __init__(
        self,
        name: str,
        on_conflict="rename_new",
        rename_template="{name} {counter}{extension}",
    ) -> None:
        if on_conflict not in CONFLICT_OPTIONS:
            raise ValueError(
                "on_conflict must be one of %s" % ", ".join(CONFLICT_OPTIONS)
            )

        self.new_name = Template.from_string(name)
        self.conflict_mode = on_conflict
        self.rename_template = Template.from_string(rename_template)

    def pipeline(self, args: dict, simulate: bool):
        fs = args["fs"]  # type: FS
        src_path = args["fs_path"]

        new_name = self.new_name.render(**args)
        if "/" in new_name:
            raise ValueError(
                "The new name cannot contain slashes. "
                "To move files or folders use `move`."
            )

        dst_path = path.join(path.dirname(src_path), new_name)

        if dst_path == src_path:
            self.print("Name did not change")
        else:
            move_action: Callable[[FS, str, FS, str], None]
            if fs.isdir(src_path):
                move_action = move_dir
            elif fs.isfile(src_path):
                move_action = move_file

            # check for conflicts
            skip, dst_path = check_conflict(
                src_fs=fs,
                src_path=src_path,
                dst_fs=fs,
                dst_path=dst_path,
                conflict_mode=self.conflict_mode,
                rename_template=self.rename_template,
                simulate=simulate,
                print=self.print,
            )

            if not skip:
                self.print("Rename to %s" % safe_description(fs, dst_path))
                if not simulate:
                    move_action(fs, src_path, fs, dst_path)

        # the next action should work with the renamed file
        return {
            "fs": fs,
            "fs_path": "./" + dst_path,
        }

    def __str__(self) -> str:
        return "Rename(new_name=%s, conflict_mode=%s)" % (
            self.new_name,
            self.conflict_mode,
        )



import sys
from typing import Union, Iterable

import simplematch as sm

from organize.utils import flatten

from .filter import Filter, FilterResult


def list_tags(path):
    import macos_tags

    tags = macos_tags.get_all(path)
    return ["{} ({})".format(tag.name, tag.color.name.lower()) for tag in tags]


class MacOSTags(Filter):
    """Filter by macOS tags

    Args:
        *tags (list(str) or str):
            The tags to filter by
    """

    name = "macos_tags"
    schema_support_instance_without_args = True

    def __init__(self, *tags) -> None:
        self.filter_tags = list(flatten(list(tags)))

    def matches(self, tags: Iterable[str]) -> Union[bool, str]:
        if not self.filter_tags:
            return True
        if not tags:
            return False
        for tag in tags:
            if any(sm.test(filter_tag, tag) for filter_tag in self.filter_tags):
                return True
        return False

    def pipeline(self, args: dict) -> FilterResult:
        fs = args["fs"]
        fs_path = args["fs_path"]
        path = fs.getsyspath(fs_path)

        if sys.platform != "darwin":
            raise EnvironmentError("The macos_tags filter is only available on macOS")

        tags = list_tags(path)

        return FilterResult(
            matches=bool(self.matches(tags)),
            updates={self.get_name(): tags},
        )

    def __str__(self):
        return "MacOSTags(tags=%s)" % ", ".join(self.filter_tags)



from unittest.mock import call

from conftest import make_files, read_files

from organize import core
from organize.filters import Python


def test_basic():
    p = Python(
        """
        return "some-string"
        """
    )
    result = p.run()
    assert result.matches
    assert result.updates == {"python": "some-string"}


def test_python(testfs, mock_echo):
    make_files(
        testfs,
        ["student-01.jpg", "student-01.txt", "student-02.txt", "student-03.txt"],
    )
    config = """
        rules:
        - locations: "."
          filters:
            - name
            - extension: txt
            - python: |
                return int(name.split('.')[0][-2:]) * 100
          actions:
            - echo: '{python}'
        """
    core.run(config, simulate=False, working_dir=testfs)
    mock_echo.assert_has_calls(
        (
            call("100"),
            call("200"),
            call("300"),
        ),
        any_order=True,
    )


def test_odd_detector(testfs, mock_echo):
    make_files(
        testfs,
        ["student-01.txt", "student-02.txt", "student-03.txt", "student-04.txt"],
    )
    config = """
        rules:
        - locations: "."
          filters:
            - name
            - python: |
                return int(name.split('-')[1]) % 2 == 1
          actions:
            - echo: 'Odd student numbers: {name}'
        """
    core.run(config, simulate=False, working_dir=testfs)
    mock_echo.assert_has_calls(
        (
            call("Odd student numbers: student-01"),
            call("Odd student numbers: student-03"),
        ),
        any_order=True,
    )


def test_python_dict(testfs, mock_echo):
    make_files(
        testfs,
        ["foo-01.jpg", "foo-01.txt", "bar-02.txt", "baz-03.txt"],
    )
    config = """
        rules:
          - locations: "."
            filters:
            - extension: txt
            - name
            - python: |
                return {
                    "name": name[:3],
                    "code": int(name.split('.')[0][-2:]) * 100,
                }
            actions:
            - echo: '{python.code} {python.name}'
        """
    core.run(config, simulate=False, working_dir=testfs)
    mock_echo.assert_has_calls(
        (
            call("100 foo"),
            call("200 bar"),
            call("300 baz"),
        ),
        any_order=True,
    )


def test_name_reverser(testfs):
    make_files(testfs, ["desrever.jpg", "emanelif.txt"])
    config = """
        rules:
          - locations: "."
            filters:
            - extension
            - name
            - python: |
                return {
                    "reversed_name": name[::-1],
                }
            actions:
            - rename: '{python.reversed_name}.{extension}'
        """
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {
        "reversed.jpg": "",
        "filename.txt": "",
    }


def test_folder_instructions(testfs):
    """
    I would like to include path/folder-instructions into the filename because I have a
    lot of different files (and there are always new categories added) I don't want
    create rules for. For example my filename is

    '2019_Jobs_CategoryA_TagB_A-Media_content-name_V01_draft_eng.docx'

    which means: Move the file to the folder '2019/Jobs/CategoryA/TagB/Media/drafts/eng'
    whereby 'A-' is an additional instruction and should be removed from the filename
    afterwards ('2019_Jobs_CategoryA_TagB_content-name_V01_draft_eng.docx').

    I have a rough idea to figure it out with python but I'm new to it (see below a
    sketch). Is there a possibility to use such variables, conditions etc. with
    organizer natively? If no, is it possible to do it with Python in Organizer at all?

     - Transform file-string into array
     - Search for 'A-...', 'V...' and 'content-name' and get index of values
     - remove value 'A-... and 'content-name' of array
     - build new filename string
     - remove value 'V...' and 'A-' of array
     - build folder-path string (convert _ to /) etc.

    """
    # inspired by: https://github.com/tfeldmann/organize/issues/52
    make_files(
        testfs,
        [
            "2019_Jobs_CategoryA_TagB_A-Media_content-name_V01_draft_eng.docx",
            "2019_Work_CategoryC_V-Test_A-Audio_V14_final.pdf",
            "other.pdf",
        ],
    )
    config = r"""
        rules:
            - locations: .
              filters:
                - extension:
                    - pdf
                    - docx
                - name:
                    contains: "_"
                - python: |
                    import fs
                    parts = []
                    instructions = dict()
                    for part in name.split("_"):
                        if part.startswith("A-"):
                            instructions["A"] = part[2:]
                        elif part.startswith("V-"):
                            instructions["V"] = part[2:]
                        elif part.startswith("content-name"):
                            instructions["content"] = part[12:]
                        else:
                            parts.append(part)
                    return {
                        "new_path": fs.path.join(*parts),
                        "instructions": instructions,
                    }
              actions:
                - echo: "New path: {python.new_path}"
                - echo: "Instructions: {python.instructions}"
                - echo: "Value of A: {python.instructions.A}"
                - move: "{python.new_path}/{name}.{extension}"
        """
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {
        "other.pdf": "",
        "2019": {
            "Jobs": {
                "CategoryA": {
                    "TagB": {
                        "V01": {
                            "draft": {
                                "eng": {
                                    "2019_Jobs_CategoryA_TagB_A-Media_content-name_V01_draft_eng.docx": "",
                                }
                            }
                        }
                    }
                }
            },
            "Work": {
                "CategoryC": {
                    "V14": {
                        "final": {
                            "2019_Work_CategoryC_V-Test_A-Audio_V14_final.pdf": "",
                        }
                    }
                }
            },
        },
    }



import logging
import re
from typing import Any

from fs.base import FS
from fs.errors import NoSysPath

from .filter import Filter, FilterResult

logger = logging.getLogger(__name__)


class FileContent(Filter):
    """Matches file content with the given regular expression

    Args:
        expr (str): The regular expression to be matched.

    Any named groups (`(?P<groupname>.*)`) in your regular expression will
    be returned like this:

    **Returns:**

    - `{filecontent.groupname}`: The text matched with the named group
      `(?P<groupname>)`
    """

    name = "filecontent"
    schema_support_instance_without_args = True

    def __init__(self, expr="(?P<all>.*)") -> None:
        self.expr = re.compile(expr, re.MULTILINE | re.DOTALL)

    def matches(self, path: str) -> Any:
        try:
            import textract

            content = textract.process(
                str(path),
                errors="ignore",
            )
            match = self.expr.search(content.decode("utf-8", errors="ignore"))
            return match
        except ImportError as e:
            raise ImportError(
                "textract is not installed. "
                "Install with pip install organize-tool[textract]"
            ) from e
        except textract.exceptions.CommandLineError as e:
            pass

    def pipeline(self, args: dict) -> FilterResult:
        fs = args["fs"]  # type: FS
        fs_path = args["fs_path"]
        if fs.isdir(fs_path):
            raise ValueError("Dirs not supported")
        try:
            syspath = fs.getsyspath(fs_path)
        except NoSysPath as e:
            raise EnvironmentError(
                "filecontent only supports the local filesystem"
            ) from e
        match = self.matches(path=syspath)
        if match:
            updates = {self.get_name(): match.groupdict()}
        else:
            updates = {}
        return FilterResult(
            matches=bool(match),
            updates=updates,
        )



import logging
import textwrap
from typing import Any, Dict, Iterable
from typing import Optional as tyOptional

from schema import Optional, Or

from .action import Action

logger = logging.getLogger(__name__)


class Python(Action):

    """Execute python code.

    Args:
        code (str): The python code to execute.
        run_in_simulation (bool):
            Whether to execute this code in simulation mode (Default false).
    """

    name = "python"
    arg_schema = Or(
        str,
        {
            "code": str,
            Optional("run_in_simulation"): bool,
        },
    )

    def __init__(self, code, run_in_simulation=False) -> None:
        self.code = textwrap.dedent(code)
        self.run_in_simulation = run_in_simulation

    def usercode(self, *args, **kwargs):
        pass  # will be overwritten by `create_method`

    def create_method(self, name: str, argnames: Iterable[str], code: str) -> None:
        globals_ = globals().copy()
        globals_["print"] = self.print
        locals_ = locals().copy()
        funccode = "def {fnc}__({arg}):\n{cod}\n\nself.{fnc} = {fnc}__\n".format(
            fnc=name,
            arg=", ".join(argnames),
            cod=textwrap.indent(textwrap.dedent(code), " " * 4),
        )
        exec(funccode, globals_, locals_)  # pylint: disable=exec-used

    def pipeline(self, args: dict, simulate: bool) -> tyOptional[Dict[str, Any]]:
        if simulate and not self.run_in_simulation:
            self.print("** Code not run in simulation. **")
            return None

        logger.info('Executing python:\n"""\n%s\n"""', self.code)
        self.create_method(name="usercode", argnames=args.keys(), code=self.code)
        self.print("Running python script.")

        result = self.usercode(**args)  # pylint: disable=assignment-from-no-return
        return result



import fs
from conftest import make_files, read_files

from organize import config, core


def test_codepost_usecase():
    files = {
        "files": {
            "Devonte-Betts.txt": "",
            "Alaina-Cornish.txt": "",
            "Dimitri-Bean.txt": "",
            "Lowri-Frey.txt": "",
            "Someunknown-User.txt": "",
        }
    }

    with fs.open_fs("temp://") as mem:
        make_files(mem, files)

        filters = config.load_from_string(
            """
            - extension: txt
            - regex: (?P<firstname>\w+)-(?P<lastname>\w+)\..*
            - python: |
                emails = {
                    "Betts": "dbetts@mail.de",
                    "Cornish": "acornish@google.com",
                    "Bean": "dbean@aol.com",
                    "Frey": "l-frey@frey.org",
                }
                if regex["lastname"] in emails:
                    return {"mail": emails[regex["lastname"]]}
            """
        )
        conf = {
            "rules": [
                {
                    "locations": [
                        {"path": "files", "filesystem": mem},
                    ],
                    "filters": filters,
                    "actions": [
                        {"move": {"dest": "files/{python.mail}.txt", "filesystem": mem}}
                    ],
                },
            ]
        }
        core.run(conf, simulate=False)
        result = read_files(mem)
        mem.tree()

    assert result == {
        "files": {
            "dbetts@mail.de.txt": "",
            "acornish@google.com.txt": "",
            "dbean@aol.com.txt": "",
            "l-frey@frey.org.txt": "",
            "Someunknown-User.txt": "",
        }
    }



import re

from fs.path import basename

from .filter import Filter, FilterResult


class Regex(Filter):

    """Matches filenames with the given regular expression

    Args:
        expr (str): The regular expression to be matched.

    **Returns:**

    Any named groups in your regular expression will be returned like this:

    - `{regex.groupname}`: The text matched with the named
      group `(?P<groupname>.*)`

    """

    name = "regex"

    arg_schema = str

    def __init__(self, expr) -> None:
        self.expr = re.compile(expr, flags=re.UNICODE)

    def matches(self, path: str):
        return self.expr.search(path)

    def pipeline(self, args: dict) -> FilterResult:
        fs_path = args["fs_path"]
        match = self.matches(basename(fs_path))
        return FilterResult(
            matches=bool(match),
            updates={
                self.get_name(): match.groupdict() if match else "",
            },
        )



import pytest
from conftest import ORGANIZE_DIR
from fs import copy

from organize.core import run

RESOURCE_DIR = str(ORGANIZE_DIR / "tests" / "resources" / "images-with-exif")


@pytest.fixture
def images_folder(testfs):
    copy.copy_dir(RESOURCE_DIR, "/", testfs, "/")
    yield testfs


def test_exif_read_camera(images_folder):
    config = """
      rules:
        - locations: "/"
          filters:
            - name
            - exif
          actions:
            - write:
                text: '{name}: {exif.image.model}'
                path: "out.txt"
                mode: append
    """
    run(config, simulate=False, working_dir=images_folder)
    out = images_folder.readtext("out.txt")
    assert "1: DMC-GX80" in out
    assert "2: NIKON D3200" in out
    assert "3: iPhone 6s" in out
    assert "4: iPhone 6s" in out
    assert "5: iPhone 5s" in out


def test_exif_filter_by_cam(images_folder):
    config = """
      rules:
        - locations: "/"
          filters:
            - name
            - exif:
                image.model: Nikon D3200
          actions:
            - write:
                text: '{name}: {exif.image.model}'
                path: "out.txt"
                mode: append
    """
    run(config, simulate=False, working_dir=images_folder)
    out = images_folder.readtext("out.txt")
    assert out.strip() == "2: NIKON D3200"


def test_exif_filter_tag_exists(images_folder):
    config = """
      rules:
        - locations: "/"
          filters:
            - name
            - exif:
                gps.gpsdate
          actions:
            - write:
                text: '{name}: {exif.gps.gpsdate}'
                path: "out.txt"
    """
    run(config, simulate=False, working_dir=images_folder)
    out = images_folder.readtext("out.txt")
    assert len(out.splitlines()) == 3
    assert "4: 2018:02:22" in out
    assert "5: 2015:07:08" in out
    assert "3: 2017:08:12" in out


def test_exif_filter_by_multiple_keys(images_folder):
    config = """
      rules:
        - locations: "/"
          filters:
            - name
            - exif:
                image.make: Apple
                exif.lensmodel: "iPhone 6s back camera 4.15mm f/2.2"
          actions:
            - move: "chosen/"
    """
    run(config, simulate=False, working_dir=images_folder)
    chosen = set(images_folder.listdir("chosen"))
    assert chosen == set(["3.jpg", "4.jpg"])



import subprocess
import sys
from datetime import datetime

from fs.base import FS

from ._timefilter import TimeFilter


class DateAdded(TimeFilter):

    """Matches files by the time the file was added to a folder.

    **`date_added` is only available on macOS!**

    Args:
        years (int): specify number of years
        months (int): specify number of months
        weeks (float): specify number of weeks
        days (float): specify number of days
        hours (float): specify number of hours
        minutes (float): specify number of minutes
        seconds (float): specify number of seconds
        mode (str):
            either 'older' or 'newer'. 'older' matches files / folders last modified before
            the given time, 'newer' matches files / folders last modified within the given
            time. (default = 'older')

    Returns:
        {date_added}: The datetime the files / folders were added.
    """

    name = "date_added"

    def get_datetime(self, args: dict) -> datetime:
        if sys.platform != "darwin":
            raise EnvironmentError("date_added is only available on macOS")

        fs = args["fs"]  # type: FS
        fs_path = args["fs_path"]

        cmd = ["mdls", "-name", "kMDItemDateAdded", "-raw", fs.getsyspath(fs_path)]
        out = subprocess.check_output(cmd, encoding="utf-8").strip()
        dt = datetime.strptime(out, "%Y-%m-%d %H:%M:%S %z")
        return dt

    def __str__(self):
        return "[DateAdded] All files / folders added %s than %s" % (
            self._mode,
            self.timedelta,
        )



"""
organize

The file management automation tool.
"""
import os
import sys
import textwrap
from typing import Optional, Tuple

import click
import fs

from . import console
from .__version__ import __version__
from .migration import NeedsMigrationError

DOCS_RTD = "https://organize.readthedocs.io"
DOCS_GHPAGES = "https://tfeldmann.github.io/organize/"

DEFAULT_CONFIG_FS_URL = "userconf://organize::/config.yaml"


def path_split(path_or_url: str) -> Tuple[str, str]:
    if sys.platform.startswith("win"):
        path_or_url = path_or_url.replace("\\", "/")
    dirname, filename = fs.path.split(path_or_url)
    return dirname, filename


def ensure_default_config():
    """
    Ensures a configuration file exists in the default location.
    """
    DEFAULT_CONFIG_TEXT = textwrap.dedent(
        """\
        # organize configuration file
        # {docs}

        rules:
          - name: "The name of this rule"
            locations:
              - # your locations here
            filters:
              - # your filters here
            actions:
              - # your actions here
        """
    ).format(docs=DOCS_RTD)

    dirname, filename = path_split(DEFAULT_CONFIG_FS_URL)
    if not filename:
        raise ValueError("invalid config path, missing filename")
    with fs.open_fs(dirname, create=True, writeable=True) as confdir:
        if not confdir.exists(filename):
            confdir.writetext(filename, DEFAULT_CONFIG_TEXT)


def config_path(fs_url: Optional[str] = None) -> Tuple[bool, str]:
    """
    Return the config path, resolved into a syspath if possible.
    If no fs_url is given, the default locations are checked.
    As last resort, a config is created in the default location.
    """
    is_syspath = False
    if not fs_url:
        # first check whether the user set a env var
        env_fs_url = os.getenv("ORGANIZE_CONFIG")
        if env_fs_url:
            fs_url = env_fs_url
        else:
            # if no env variable is given we make sure that there is a config file in
            # the default location
            ensure_default_config()
            fs_url = DEFAULT_CONFIG_FS_URL

    dirname, filename = path_split(fs_url)
    try:
        with fs.open_fs(dirname) as confdir:
            config_path = confdir.getsyspath(filename)
            is_syspath = True
    except Exception:
        config_path = fs_url
    return is_syspath, config_path


def read_config(fs_url: Optional[str] = None) -> Tuple[str, str]:
    """
    Read the config at the given fs_url.
    If no fs_url is given, try the default locations.
    """
    _, fs_url = config_path(fs_url)
    dirname, filename = path_split(fs_url)
    with fs.open_fs(dirname) as confdir:
        return (fs_url, confdir.readtext(filename))


class NaturalOrderGroup(click.Group):
    def list_commands(self, ctx):
        return self.commands.keys()


class TagType(click.ParamType):
    name = "tag"

    def convert(self, value, param, ctx):
        if not value:
            return tuple()
        return tuple(tag.strip() for tag in value.split(","))


CLI_CONFIG = click.argument(
    "config",
    required=False,
    type=str,
)
CLI_WORKING_DIR_OPTION = click.option(
    "--working-dir",
    default="",
    type=str,
    help="The working directory",
)
CLI_TAGS = click.option(
    "--tags",
    type=TagType(),
    default="",
    help="tags to run",
)
CLI_SKIP_TAGS = click.option(
    "--skip-tags",
    type=TagType(),
    default="",
    help="tags to skip",
)


def execute(
    config: Optional[str],
    working_dir: str,
    simulate: bool,
    tags: Optional[Tuple[str]] = None,
    skip_tags: Optional[Tuple[str]] = None,
):
    from schema import SchemaError
    from . import core

    config_path, config_text = read_config(config)

    try:
        console.info(config=config_path, working_dir=working_dir)
        core.run(
            rules=config_text,
            simulate=simulate,
            working_dir=working_dir,
            tags=tags,
            skip_tags=skip_tags,
        )
    except NeedsMigrationError as e:
        from .migration import MIGRATION_DOCS_URL

        console.error(e, title="Config needs migration")
        console.warn(
            "Your config file needs some updates to work with organize v2.\n"
            "Please see the migration guide at\n\n"
            "%s" % MIGRATION_DOCS_URL
        )
        sys.exit(1)
    except SchemaError as e:
        console.error("Invalid config file!")
        for err in e.autos:
            if err and len(err) < 200:
                core.highlighted_console.print(err)
    except Exception as e:
        core.highlighted_console.print_exception()
    except (EOFError, KeyboardInterrupt):
        console.status.stop()
        console.warn("Aborted")


@click.group(
    help=__doc__,
    cls=NaturalOrderGroup,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.version_option(__version__)
def cli():
    pass


@cli.command()
@CLI_CONFIG
@CLI_WORKING_DIR_OPTION
@CLI_TAGS
@CLI_SKIP_TAGS
def run(config: Optional[str], working_dir: str, tags, skip_tags):
    """Organizes your files according to your rules."""
    execute(
        config=config,
        working_dir=working_dir,
        simulate=False,
        tags=tags,
        skip_tags=skip_tags,
    )


@cli.command()
@CLI_CONFIG
@CLI_WORKING_DIR_OPTION
@CLI_TAGS
@CLI_SKIP_TAGS
def sim(config: Optional[str], working_dir: str, tags, skip_tags):
    """Simulates a run (does not touch your files)."""
    execute(
        config=config,
        working_dir=working_dir,
        simulate=True,
        tags=tags,
        skip_tags=skip_tags,
    )


@cli.command()
@click.argument("config", required=False, type=str)
@click.option(
    "--editor",
    envvar="EDITOR",
    help="The editor to use. (Default: $EDITOR)",
)
def edit(config: Optional[str], editor):
    """Edit the rules.

    If called without arguments it will open the default config file in $EDITOR.
    """
    is_syspath, confpath = config_path(config)
    if is_syspath:
        click.edit(filename=confpath, editor=editor)
    else:
        click.echo("Not a local config path: %s" % confpath)


@cli.command()
@CLI_CONFIG
@click.option("--debug", is_flag=True, help="Verbose output")
def check(config: str, debug):
    """Checks whether a given config file is valid.

    If called without arguments it will check the default config file.
    """
    from . import migration
    from .config import cleanup, load_from_string, validate
    from .core import highlighted_console as out
    from .core import replace_with_instances

    try:
        config_path, config_str = read_config(config)
        print("Checking: %s" % config_path)

        if debug:
            out.rule("Raw", align="left")
            out.print(config_str)

        rules = load_from_string(config_str)

        if debug:
            out.print("\n\n")
            out.rule("Loaded", align="left")
            out.print(rules)

        rules = cleanup(rules)

        if debug:
            out.print("\n\n")
            out.rule("Cleaned", align="left")
            out.print(rules)

        if debug:
            out.print("\n\n")
            out.rule("Migration from v1", align="left")

        migration.migrate_v1(rules)

        if debug:
            out.print("Not needed.")
            out.print("\n\n")
            out.rule("Schema validation", align="left")

        validate(rules)

        if debug:
            out.print("Validation ok.")
            out.print("\n\n")
            out.rule("Instantiation", align="left")

        # warnings = replace_with_instances(rules)
        # if debug:
        #     out.print(rules)
        #     for msg in warnings:
        #         out.print("Warning: %s" % msg)

        if debug:
            out.print("\n\n")
            out.rule("Result", align="left")
        out.print("Config is valid.")

    except Exception as e:
        out.print_exception()
        sys.exit(1)


@cli.command()
@CLI_CONFIG
@click.option("--path", is_flag=True, help="Print the path instead of revealing it.")
def reveal(config: Optional[str], path: bool):
    """Reveals the default config file."""
    is_syspath, confpath = config_path(config)
    if path:
        click.echo(confpath)
        return
    try:
        # convert the url
        dirname, _ = path_split(confpath)
        import webbrowser

        with fs.open_fs(dirname) as dirfs:
            dir_url = dirfs.geturl("/")
            if not is_syspath:
                raise ValueError("not a local path")
            webbrowser.open(dir_url)
    except Exception as e:
        click.echo("Cannot reveal this config (%s)" % e)
        click.echo(confpath)


@cli.command()
def schema():
    """Prints the json schema for config files."""
    # ORGANIZE SCHEMA IS DEPRECATED AND WILL BE REMOVED IN THE FUTURE
    import json

    from .config import CONFIG_SCHEMA
    from .console import console as richconsole

    js = json.dumps(
        CONFIG_SCHEMA.json_schema(
            schema_id="https://tfeldmann.de/organize.schema.json",
        )
    )
    richconsole.print_json(js)


@cli.command()
def docs():
    """Opens the documentation."""
    import webbrowser

    webbrowser.open(DOCS_RTD)


if __name__ == "__main__":
    cli()



import fs
from unittest.mock import patch

from organize.actions import Trash
from organize.core import run


def test_trash_mocked():
    with patch("send2trash.send2trash") as mck:
        trash = Trash()
        trash.trash(path="~/Desktop/Test.zip", simulate=False)
        mck.assert_called_with("~/Desktop/Test.zip")


def test_trash_file():
    config = """
    rules:
      - locations: "."
        filters:
          - name: "test-trash-209318123"
        actions:
          - trash
    """
    with fs.open_fs("temp://") as tmp:
        tmp.writetext("test-trash-209318123.txt", "Content")
        assert tmp.exists("test-trash-209318123.txt")
        run(config, simulate=False, working_dir=tmp.getsyspath("/"))
        assert not tmp.exists("test-trash-209318123.txt")


def test_trash_folder():
    config = """
    rules:
      - locations: "."
        targets: dirs
        filters:
          - name: "test-trash-209318123"
        actions:
          - trash
    """
    with fs.open_fs("temp://") as tmp:
        tmp_dir = tmp.makedir("test-trash-209318123")
        tmp_dir.touch("file.txt")
        assert tmp.exists("test-trash-209318123/file.txt")
        run(config, simulate=False, working_dir=tmp.getsyspath("/"))
        assert not tmp.exists("test-trash-209318123/file.txt")
        assert not tmp.exists("test-trash-209318123")



from typing import Dict, Type

from .action import Action
from .confirm import Confirm
from .copy import Copy
from .delete import Delete
from .echo import Echo
from .macos_tags import MacOSTags
from .move import Move
from .python import Python
from .rename import Rename
from .shell import Shell
from .symlink import Symlink
from .trash import Trash
from .write import Write

ACTIONS = {
    Confirm.name: Confirm,
    Copy.name: Copy,
    Delete.name: Delete,
    Echo.name: Echo,
    MacOSTags.name: MacOSTags,
    Move.name: Move,
    Python.name: Python,
    Rename.name: Rename,
    Shell.name: Shell,
    Symlink.name: Symlink,
    Trash.name: Trash,
    Write.name: Write,
}  # type: Dict[str, Type[Action]]



import logging
import os

from fs import path
from fs.base import FS
from fs.osfs import OSFS

from organize.utils import Template

from .action import Action

logger = logging.getLogger(__name__)


class Symlink(Action):

    """Create a symbolic link.

    Args:
        dest (str):
            The symlink destination. If **dest** ends with a slash `/``, create the
            symlink in the given directory. Can contain placeholders.

    Only the local filesystem is supported.
    """

    name = "symlink"

    def __init__(self, dest):
        self._dest = Template.from_string(dest)

    def pipeline(self, args: dict, simulate: bool):
        fs = args["fs"]  # type: FS
        fs_path = args["fs_path"]  # type: str

        if not isinstance(fs, OSFS):
            raise EnvironmentError("Symlinks only work on the local filesystem.")

        dest = os.path.expanduser(self._dest.render(**args))
        if dest.endswith("/"):
            dest = path.join(dest, path.basename(fs_path))

        self.print("Creating symlink: %s" % dest)
        if not simulate:
            os.makedirs(os.path.dirname(dest), exist_ok=True)
            os.symlink(fs.getsyspath(fs_path), dest)



from copy import deepcopy

import pytest
from conftest import make_files, read_files
from fs.base import FS

from organize import core

files = {
    "test.txt": "",
    "file.txt": "Hello world\nAnother line",
    "another.txt": "",
    "folder": {
        "x.txt": "",
    },
}


@pytest.fixture
def testfiles(testfs) -> FS:
    make_files(testfs, files)
    yield testfs


def test_copy_on_itself(testfiles):
    config = """
    rules:
      - locations: "/"
        actions:
          - copy: "/"
    """
    core.run(config, simulate=False, working_dir=testfiles)
    result = read_files(testfiles)
    assert result == files


def test_copy_basic(testfs):
    config = """
    rules:
      - locations: "."
        filters:
          - name: test
        actions:
          - copy: newname.pdf
    """
    make_files(testfs, ["asd.txt", "newname 2.pdf", "newname.pdf", "test.pdf"])
    core.run(config, simulate=False, working_dir=testfs)
    assert read_files(testfs) == {
        "newname.pdf": "",
        "newname 2.pdf": "",
        "newname 3.pdf": "",
        "test.pdf": "",
        "asd.txt": "",
    }


def test_copy_into_dir(testfiles):
    config = """
    rules:
      - locations: "/"
        actions:
          - copy: "a brand/new/folder/"
    """
    core.run(config, simulate=False, working_dir=testfiles)
    result = read_files(testfiles)
    assert result == {
        "test.txt": "",
        "file.txt": "Hello world\nAnother line",
        "another.txt": "",
        "folder": {
            "x.txt": "",
        },
        "a brand": {
            "new": {
                "folder": {
                    "test.txt": "",
                    "file.txt": "Hello world\nAnother line",
                    "another.txt": "",
                }
            }
        },
    }


def test_copy_into_dir_subfolders(testfiles):
    config = """
    rules:
      - locations: "/"
        subfolders: true
        actions:
          - copy: "a brand/new/folder/"
    """
    core.run(config, simulate=False, working_dir=testfiles)
    result = read_files(testfiles)
    assert result == {
        "test.txt": "",
        "file.txt": "Hello world\nAnother line",
        "another.txt": "",
        "folder": {
            "x.txt": "",
        },
        "a brand": {
            "new": {
                "folder": {
                    "test.txt": "",
                    "file.txt": "Hello world\nAnother line",
                    "another.txt": "",
                    "x.txt": "",
                }
            }
        },
    }


@pytest.mark.parametrize(
    "mode,files,test_txt_content",
    [
        ("skip", ["file.txt", "test.txt"], "old"),
        ("overwrite", ["file.txt", "test.txt"], "new"),
        ("rename_new", ["file.txt", "test.txt", "test 2.txt"], "old"),
        ("rename_existing", ["file.txt", "test.txt", "test 2.txt"], "new"),
    ],
)
def test_copy_conflict(testfs: FS, mode, files, test_txt_content):
    config = """
    rules:
      - locations: "/"
        filters:
          - name: file
        actions:
          - copy:
              dest: "test.txt"
              on_conflict: {}
    """.format(
        mode
    )
    testfs.writetext("file.txt", "new")
    testfs.writetext("test.txt", "old")
    core.run(config, simulate=False, working_dir=testfs)
    assert set(testfs.listdir("/")) == set(files)
    assert testfs.readtext("file.txt") == "new"
    assert testfs.readtext("test.txt") == test_txt_content


def test_does_not_create_folder_in_simulation(testfs):
    config = """
        rules:
          - locations: "."
            actions:
              - copy: "new-subfolder/"
              - copy: "copyhere/"
        """
    make_files(testfs, files)
    core.run(config, simulate=True, working_dir=testfs)
    result = read_files(testfs)
    assert result == files

    core.run(config, simulate=False, working_dir=testfs)
    result = read_files(testfs)

    expected = deepcopy(files)
    expected["new-subfolder"] = deepcopy(files)
    expected["new-subfolder"].pop("folder")
    expected["copyhere"] = deepcopy(files)
    expected["copyhere"].pop("folder")

    assert result == expected



from conftest import make_files, read_files

from organize import core

CONTENT_SMALL = "COPY CONTENT"
CONTENT_LARGE = "XYZ" * 300000

CONFIG_DEEP_DUP_DELETE = """
rules:
  - locations: "."
    subfolders: true
    filters:
      - duplicate:
          detect_original_by: name
    actions:
      - delete
"""


def test_duplicate_smallfiles(testfs):
    files = {
        "unique.txt": "I'm unique.",
        "unique_too.txt": "I'm unique: too.",
        "a.txt": CONTENT_SMALL,
        "copy2.txt": CONTENT_SMALL,
        "other": {
            "copy.txt": CONTENT_SMALL,
            "copy.jpg": CONTENT_SMALL,
            "large.txt": CONTENT_LARGE,
        },
        "large_unique.txt": CONTENT_LARGE,
    }

    make_files(testfs, files)
    core.run(CONFIG_DEEP_DUP_DELETE, simulate=False, working_dir=testfs)
    result = read_files(testfs)
    testfs.tree()
    assert result == {
        "unique.txt": "I'm unique.",
        "unique_too.txt": "I'm unique: too.",
        "a.txt": CONTENT_SMALL,
        "other": {
            "large.txt": CONTENT_LARGE,
        },
    }


def test_duplicate_largefiles(testfs):
    files = {
        "unique.txt": CONTENT_LARGE + "1",
        "unique_too.txt": CONTENT_LARGE + "2",
        "a.txt": CONTENT_LARGE,
        "copy2.txt": CONTENT_LARGE,
        "other": {
            "copy.txt": CONTENT_LARGE,
            "copy.jpg": CONTENT_LARGE,
            "large.txt": CONTENT_LARGE,
        },
    }

    make_files(testfs, files)
    core.run(CONFIG_DEEP_DUP_DELETE, simulate=False, working_dir=testfs)
    result = read_files(testfs)
    testfs.tree()
    assert result == {
        "unique.txt": CONTENT_LARGE + "1",
        "unique_too.txt": CONTENT_LARGE + "2",
        "a.txt": CONTENT_LARGE,
        "other": {},
    }


# TODO detect_original_by: first_seen
# TODO detect_original_by: created
# TODO detect_original_by: lastmodified



import logging

from .action import Action

logger = logging.getLogger(__name__)


class Trash(Action):

    """Move a file or dir into the trash."""

    name = "trash"

    @classmethod
    def get_schema(cls):
        return cls.name

    def trash(self, path: str, simulate: bool):
        from send2trash import send2trash

        self.print('Trash "%s"' % path)
        if not simulate:
            logger.info("Moving file %s into trash.", path)
            send2trash(path)

    def pipeline(self, args: dict, simulate: bool):
        fs = args["fs"]
        fs_path = args["fs_path"]
        self.trash(path=fs.getsyspath(fs_path), simulate=simulate)



from datetime import datetime, timedelta


def age_condition_applies(dt: datetime, age: timedelta, mode: str, reference: datetime):
    """
    Returns whether `dt` is older / newer (`mode`) than `age` as measured on `reference`
    """
    if mode not in ("older", "newer"):
        raise ValueError(mode)

    is_past = (dt + age).timestamp() < reference.timestamp()
    return (mode == "older") == is_past



from ..utils import Template
from .action import Action


class Echo(Action):

    """Prints the given message.

    This can be useful to test your rules, especially in combination with placeholder
    variables.

    Args:
        msg (str): The message to print. Accepts placeholder variables.
    """

    name = "echo"

    @classmethod
    def get_schema(cls):
        return {cls.name: str}

    def __init__(self, msg):
        self.msg = Template.from_string(msg)

    def pipeline(self, args: dict, simulate: bool) -> None:
        full_msg = self.msg.render(**args)
        self.print("%s" % full_msg)

    def __str__(self) -> str:
        return 'Echo(msg="%s")' % self.msg



from typing import Callable, Union

import jinja2
from fs import errors, open_fs
from fs.base import FS
from fs.move import move_dir, move_file
from fs.opener.errors import OpenerError
from fs.path import basename, dirname, join, splitext
from jinja2 import Template

from organize.utils import expand_args, is_same_resource, safe_description

from .trash import Trash

CONFLICT_OPTIONS = (
    "skip",
    "overwrite",
    "trash",
    "rename_new",
    "rename_existing",
    # "keep_newer",
    # "keep_older",
)


def next_free_name(
    fs: FS, template: jinja2.Template, name: str, extension: str, start=2
) -> str:
    """
    Increments {counter} in the template until the given resource does not exist.

    Args:
        fs (FS): the filesystem to work on
        template (jinja2.Template):
            A jinja2 template with placeholders for {name}, {extension} and {counter}
        name (str): The wanted filename
        extension (str): the wanted extension

    Raises:
        ValueError if no free name can be found with the given template.so

    Returns:
        (str) A filename according to the given template that does not exist on **fs**.
    """
    counter = start
    prev_candidate = ""
    if not extension.startswith("."):
        extension = ".%s" % extension
    wanted = "%s%s" % (name, extension)
    if not fs.exists(wanted):
        return wanted
    while True:
        candidate = template.render(name=name, extension=extension, counter=counter)
        if not fs.exists(candidate):
            return candidate
        if prev_candidate == candidate:
            raise ValueError(
                "Could not find a free filename for the given template. "
                'Maybe you forgot the "{counter}" placeholder?'
            )
        prev_candidate = candidate
        counter += 1


def resolve_overwrite_conflict(
    src_fs: FS,
    src_path: str,
    dst_fs: FS,
    dst_path: str,
    conflict_mode: str,
    rename_template: Template,
    simulate: bool,
    print: Callable,
) -> Union[None, str]:
    """
    Returns:
        - A new path if applicable
        - None if this action should be skipped.
    """
    if is_same_resource(src_fs, src_path, dst_fs, dst_path):
        print("Same resource: Skipped.")
        return None

    if conflict_mode == "trash":
        Trash().run(fs=dst_fs, fs_path=dst_path, simulate=simulate)
        return dst_path

    elif conflict_mode == "skip":
        print("Skipped.")
        return None

    elif conflict_mode == "overwrite":
        print("Overwrite %s." % safe_description(dst_fs, dst_path))
        if not simulate:
            if dst_fs.isdir(dst_path):
                dst_fs.removedir(dst_path)
            elif dst_fs.isfile(dst_path):
                dst_fs.remove(dst_path)
        return dst_path

    elif conflict_mode == "rename_new":
        stem, ext = splitext(dst_path)
        name = next_free_name(
            fs=dst_fs,
            name=stem,
            extension=ext,
            template=rename_template,
        )
        return name

    elif conflict_mode == "rename_existing":
        stem, ext = splitext(dst_path)
        name = next_free_name(
            fs=dst_fs,
            name=stem,
            extension=ext,
            template=rename_template,
        )
        print('Renaming existing to: "%s"' % name)
        if not simulate:
            if dst_fs.isdir(dst_path):
                move_dir(dst_fs, dst_path, dst_fs, name)
            elif dst_fs.isfile(dst_path):
                move_file(dst_fs, dst_path, dst_fs, name)
        return dst_path

    raise ValueError("Unknown conflict_mode %s" % conflict_mode)


def dst_from_options(src_path, dest, filesystem, args: dict):
    # append the original resource name if destination is a dir (ends with "/")
    dst_path = expand_args(dest, args)
    if dst_path.endswith(("\\", "/")):
        dst_path = join(dst_path, basename(src_path))

    if filesystem:
        if isinstance(filesystem, str):
            dst_fs = expand_args(filesystem, args)
        else:
            dst_fs = filesystem
    else:
        dst_fs = dirname(dst_path)
        dst_path = basename(dst_path)
    return dst_fs, dst_path


def check_conflict(
    src_fs: FS,
    src_path: str,
    dst_fs: FS,
    dst_path: str,
    conflict_mode: str,
    rename_template: Template,
    simulate: bool,
    print: Callable,
):
    skip = False
    try:
        check_fs = open_fs(dst_fs, create=False, writeable=True)
        if check_fs.exists(dst_path):
            print(
                '%s already exists! (conflict mode is "%s").'
                % (safe_description(dst_fs, dst_path), conflict_mode)
            )
            new_path = resolve_overwrite_conflict(
                src_fs=src_fs,
                src_path=src_path,
                dst_fs=check_fs,
                dst_path=dst_path,
                conflict_mode=conflict_mode,
                rename_template=rename_template,
                simulate=simulate,
                print=print,
            )
            if new_path is not None:
                dst_path = new_path
            else:
                skip = True
    except (errors.CreateFailed, OpenerError):
        pass

    return skip, dst_path



from typing import Union

from fs.base import FS

from organize.utils import flatten

from .filter import Filter, FilterResult


class Extension(Filter):
    """Filter by file extension

    Args:
        *extensions (list(str) or str):
            The file extensions to match (does not need to start with a colon).

    **Returns:**

    - `{extension}`: the original file extension (without colon)
    """

    name = "extension"
    schema_support_instance_without_args = True

    def __init__(self, *extensions) -> None:
        self.extensions = list(map(self.normalize_extension, flatten(list(extensions))))

    @staticmethod
    def normalize_extension(ext: str) -> str:
        """strip colon and convert to lowercase"""
        if ext.startswith("."):
            return ext[1:].lower()
        else:
            return ext.lower()

    def matches(self, ext: str) -> Union[bool, str]:
        if not self.extensions:
            return True
        if not ext:
            return False
        return self.normalize_extension(ext) in self.extensions

    def pipeline(self, args: dict) -> FilterResult:
        fs = args["fs"]  # type: FS
        fs_path = args["fs_path"]
        if fs.isdir(fs_path):
            raise ValueError("Dirs not supported")

        # suffix is the extension with dot
        suffix = fs.getinfo(fs_path).suffix
        ext = suffix[1:]
        return FilterResult(
            matches=bool(self.matches(ext)),
            updates={self.get_name(): ext},
        )

    def __str__(self):
        return "Extension(%s)" % ", ".join(self.extensions)



from pathlib import Path
from unittest.mock import patch

from organize.actions import Shell


def test_shell_basic():
    shell = Shell(
        "echo 'Hello World'",
        simulation_output="-sim-",
        simulation_returncode=127,
    )
    result = shell.run(simulate=True)
    assert "-sim-" in result["shell"]["output"]
    assert 127 == result["shell"]["returncode"]

    result = shell.run(simulate=False)
    result = result["shell"]
    assert "Hello World" in result["output"]  # windows escapes the string
    assert result["returncode"] == 0


def test_shell_template_simulation():
    shell = Shell("echo '{msg}'", run_in_simulation=True)
    result = shell.run(msg="Hello", simulate=True)
    result = result["shell"]
    assert "Hello" in result["output"]
    assert result["returncode"] == 0
